import logging
import os
import resource
import subprocess
import sys
import errno

from machetli import tools


# This function is copied from lab.calls.call (<https://lab.readthedocs.org>).
def set_limit(kind, soft_limit, hard_limit):
    try:
        resource.setrlimit(kind, (soft_limit, hard_limit))
    except (OSError, ValueError) as err:
        logging.error(
            f"Resource limit for {kind} could not be set to "
            f"[{soft_limit}, {hard_limit}] ({err})"
        )


class Run:
    """Define an executable command with time and memory limits.
    """

    def __init__(self, command, time_limit=1800, memory_limit=None, log_output=None):
        """*command* is a list of strings that starts your program with
        the desired parameters on a Linux machine.

        After *time_limit* seconds, the subprocess of *command* 
        is killed.

        Above a memory usage of *memory_limit* MiB, the subprocess of
        *command* is killed.

        Use the *log_output* option ``"on_fail"`` if you want log files to be
        written when *command* terminates on a non-zero exit code or use the
        option ``"always"`` if you want them always to be written. These options
        only work in combination with the :func:`machetli.run.run_all` function.
        """
        self.command = command
        self.time_limit = time_limit
        self.memory_limit = memory_limit
        self.log_on_fail = log_output == "on_fail"
        self.log_always = log_output == "always"

    def __repr__(self):
        return f'Run(\"{" ".join([os.path.basename(part) for part in self.command])}\")'

    def start(self, state):
        """Format the command with the entries of *state* and execute it with
        `subprocess.Popen <https://docs.python.org/3/library/subprocess.html#subprocess.Popen>`_.
        Return the 3-tuple (stdout, stderr, returncode) with the values obtained 
        from the executed command.
        """
        # These declarations are needed for the _prepare_call() function.
        time_limit = self.time_limit
        memory_limit = self.memory_limit

        def _prepare_call():
            # When the soft time limit is reached, SIGXCPU is emitted. Once we
            # reach the higher hard time limit, SIGKILL is sent. Having some
            # padding between the two limits allows programs to handle SIGXCPU.
            if time_limit is not None:
                set_limit(resource.RLIMIT_CPU, time_limit, time_limit + 5)
            if memory_limit is not None:
                _, hard_mem_limit = resource.getrlimit(resource.RLIMIT_AS)
                # Convert memory from MiB to Bytes.
                set_limit(resource.RLIMIT_AS, memory_limit *
                          1024 * 1024, hard_mem_limit)
            set_limit(resource.RLIMIT_CORE, 0, 0)

        formatted_command = [part.format(**state) for part in self.command]
        logging.debug(f"Formatted command:\n{formatted_command}")

        cwd = state["cwd"] if "cwd" in state else None

        try:
            process = subprocess.Popen(formatted_command,
                                       preexec_fn=_prepare_call,
                                       stdout=subprocess.PIPE,
                                       stderr=subprocess.PIPE,
                                       text=True,
                                       cwd=cwd)
        except OSError as err:
            if err.errno == errno.ENOENT:
                sys.exit('Error: Call "{}" failed. One of the files was not found.'.format(
                    ' '.join(formatted_command)))
            else:
                raise

        out_str, err_str = process.communicate()

        return (out_str, err_str, process.returncode)


class RunWithInputFile(Run):
    """Extension of the :class:`Run <machetli.run.Run>` class adding
    the option of sending the content of a file to stdin.
    """
    # e.g., in a command like ``path/to/./my_executable < my_input_file``.

    def __init__(self, command, input_file, **kwargs):
        """*input_file* is the path to the file whose content should be sent to
        the stdin of the executed *command*.
        """
        super().__init__(command, **kwargs)
        self.input_file = input_file

    def start(self, state):
        """Same as the :meth:`base method <machetli.run.Run.start>`, with
        the addition of the content from *input_file* being passed to the
        stdin of the executed *command*.
        """
        # These declarations are needed for the _prepare_call() function.
        time_limit = self.time_limit
        memory_limit = self.memory_limit

        def _prepare_call():
            # When the soft time limit is reached, SIGXCPU is emitted. Once we
            # reach the higher hard time limit, SIGKILL is sent. Having some
            # padding between the two limits allows programs to handle SIGXCPU.
            if time_limit is not None:
                set_limit(resource.RLIMIT_CPU, time_limit, time_limit + 5)
            if memory_limit is not None:
                _, hard_mem_limit = resource.getrlimit(resource.RLIMIT_AS)
                # Convert memory from MiB to Bytes.
                set_limit(resource.RLIMIT_AS, memory_limit *
                          1024 * 1024, hard_mem_limit)
            set_limit(resource.RLIMIT_CORE, 0, 0)

        formatted_command = [part.format(**state) for part in self.command]

        cwd = state["cwd"] if "cwd" in state else None

        try:
            process = subprocess.Popen(formatted_command,
                                       preexec_fn=_prepare_call,
                                       stdout=subprocess.PIPE,
                                       stderr=subprocess.PIPE,
                                       stdin=subprocess.PIPE,
                                       text=True,
                                       cwd=cwd)
        except OSError as err:
            if err.errno == errno.ENOENT:
                sys.exit('Error: Call "{}" failed. One of the files was not found.'.format(
                    ' '.join(formatted_command)))
            else:
                raise

        with open(self.input_file.format(**state), "r") as file:
            input_text = file.read()

        out_str, err_str = process.communicate(input=input_text)

        return (out_str, err_str, process.returncode)


def run_all(state):
    """Start all runs in *state["runs"]* and return a *results* dictionary
    where run outputs of run *run_name* can be accessed via:
    
    - *results[run_name]["stdout"]*,
    - *results[run_name]["stderr"]* and
    - *results[run_name]["returncode"]*.
    """
    assert "runs" in state, "Could not find entry \"runs\" in state."
    results = {}
    for name, run in state["runs"].items():
        stdout, stderr, returncode = run.start(state)
        if run.log_always or run.log_on_fail and returncode != 0:
            cwd = state["cwd"] if "cwd" in state else os.path.dirname(
                tools.get_script_path())
            if stdout:
                with open(os.path.join(cwd, f"{name}.log"), "w") as logfile:
                    logfile.write(stdout)
            if stderr:
                with open(os.path.join(cwd, f"{name}.err"), "w") as errfile:
                    errfile.write(stderr)
        results.update(
            {name: {"stdout": stdout, "stderr": stderr, "returncode": returncode}}
        )
    return results


def run_and_parse_all(state, parsers):
    """Execute :func:`machetli.run.run_all` and apply all *parsers* to the
    generated stdout and stderr outputs. Return an updated version of the
    *results* dictionary containing the parsing results in place of the actual
    stdout and stderr outputs. *parsers* can be a list of :class:`machetli.parser.Parser`
    objects or a single one.
    """
    results = run_all(state)
    parsed_results = {}
    parsers = [parsers] if not isinstance(parsers, list) else parsers
    for name, result in results.items():
        parsed_results.update(
            {name: {"stdout": {}, "stderr": {},
                    "returncode": result["returncode"]}}
        )
        for parser in parsers:
            parsed_results[name]["stdout"].update(
                parser.parse(name, result["stdout"]))
            parsed_results[name]["stderr"].update(
                parser.parse(name, result["stderr"]))
    parsed_results["raw_results"] = results
    return parsed_results



from machetli.planning.downward_lib.pddl.task_element import TaskElement


class Function(TaskElement):
    def accept(self, visitor):
        return visitor.visit_function(self)

    def __init__(self, name, arguments, type_name):
        self.name = name
        self.arguments = arguments
        if type_name != "number":
            raise SystemExit("Error: object fluents not supported\n" +
                             "(function %s has type %s)" % (name, type_name))
        self.type_name = type_name

    def __str__(self):
        result = "%s(%s)" % (self.name, ", ".join(map(str, self.arguments)))
        if self.type_name:
            result += ": %s" % self.type_name
        return result

    def dump_pddl(self, output, indent="  "):
        output.write("%s(%s " % (indent, self.name))
        for arg in self.arguments:
            output.write("%s - %s " % (arg.name, arg.type_name))
        output.write(") - %s\n" % self.type_name)



from machetli.planning.downward_lib.pddl.task_element import TaskElement

from . import axioms
from . import predicates


class Task(TaskElement):
    def accept(self, visitor):
        return visitor.visit_task(self)



    def __init__(self, domain_name, task_name, requirements,
                 types, objects, predicates, functions, init, goal,
                 actions, axioms, use_metric):
        self.domain_name = domain_name
        self.task_name = task_name
        self.requirements = requirements
        self.types = types
        self.objects = objects
        self.predicates = predicates
        self.functions = functions
        self.init = init
        self.goal = goal
        self.actions = actions
        self.axioms = axioms
        self.axiom_counter = 0
        self.use_min_cost_metric = use_metric

    def add_axiom(self, parameters, condition):
        name = "new-axiom@%d" % self.axiom_counter
        self.axiom_counter += 1
        axiom = axioms.Axiom(name, parameters, len(parameters), condition)
        self.predicates.append(predicates.Predicate(name, parameters))
        self.axioms.append(axiom)
        return axiom

    def dump(self):
        print("Problem %s: %s [%s]" % (
            self.domain_name, self.task_name, self.requirements))
        print("Types:")
        for type in self.types:
            print("  %s" % type)
        print("Objects:")
        for obj in self.objects:
            print("  %s" % obj)
        print("Predicates:")
        for pred in self.predicates:
            print("  %s" % pred)
        print("Functions:")
        for func in self.functions:
            print("  %s" % func)
        print("Init:")
        for fact in self.init:
            print("  %s" % fact)
        print("Goal:")
        self.goal.dump()
        print("Actions:")
        for action in self.actions:
            action.dump()
        if self.axioms:
            print("Axioms:")
            for axiom in self.axioms:
                axiom.dump()


class Requirements:
    def __init__(self, requirements):
        self.requirements = requirements
        for req in requirements:
            assert req in (
                ":strips", ":adl", ":typing", ":negation", ":equality",
                ":negative-preconditions", ":disjunctive-preconditions",
                ":existential-preconditions", ":universal-preconditions",
                ":quantified-preconditions", ":conditional-effects",
                ":derived-predicates", ":action-costs"), req

    def __str__(self):
        return ", ".join(self.requirements)



SAS_FILE_VERSION = 3

DEBUG = False


class SASTask:
    """Planning task in finite-domain representation.

    The user is responsible for making sure that the data fits a
    number of structural restrictions. For example, conditions should
    generally be sorted and mention each variable at most once. See
    the validate methods for details."""

    def __init__(self, variables, mutexes, init, goal,
                 operators, axioms, metric):
        self.variables = variables
        self.mutexes = mutexes
        self.init = init
        self.goal = goal
        self.operators = sorted(operators, key=lambda op: (
            op.name, op.prevail, op.pre_post))
        self.axioms = sorted(axioms, key=lambda axiom: (
            axiom.condition, axiom.effect))
        self.metric = metric
        if DEBUG:
            self.validate()

    def validate(self):
        """Fail an assertion if the task is invalid.

        A task is valid if all its components are valid. Valid tasks
        are almost in a kind of "canonical form", but not quite. For
        example, operators and axioms are permitted to be listed in
        any order, even though it would be possible to require some
        kind of canonical sorting.

        Note that we require that all derived variables are binary.
        This is stricter than what later parts of the planner are
        supposed to handle, but some parts of the translator rely on
        this. We might want to consider making this a general
        requirement throughout the planner.

        Note also that there is *no* general rule on what the init (=
        fallback) value of a derived variable is. For example, in
        PSR-Large #1, it can be either 0 or 1. While it is "usually"
        1, code should not rely on this.
        """
        self.variables.validate()
        for mutex in self.mutexes:
            mutex.validate(self.variables)
        self.init.validate(self.variables)
        self.goal.validate(self.variables)
        for op in self.operators:
            op.validate(self.variables)
        for axiom in self.axioms:
            axiom.validate(self.variables, self.init)
        assert self.metric is False or self.metric is True, self.metric

    def dump(self):
        print("variables:")
        self.variables.dump()
        print("%d mutex groups:" % len(self.mutexes))
        for mutex in self.mutexes:
            print("group:")
            mutex.dump()
        print("init:")
        self.init.dump()
        print("goal:")
        self.goal.dump()
        print("%d operators:" % len(self.operators))
        for operator in self.operators:
            operator.dump()
        print("%d axioms:" % len(self.axioms))
        for axiom in self.axioms:
            axiom.dump()
        print("metric: %s" % self.metric)

    def output(self, stream):
        print("begin_version", file=stream)
        print(SAS_FILE_VERSION, file=stream)
        print("end_version", file=stream)
        print("begin_metric", file=stream)
        print(int(self.metric), file=stream)
        print("end_metric", file=stream)
        self.variables.output(stream)
        print(len(self.mutexes), file=stream)
        for mutex in self.mutexes:
            mutex.output(stream)
        self.init.output(stream)
        self.goal.output(stream)
        print(len(self.operators), file=stream)
        for op in self.operators:
            op.output(stream)
        print(len(self.axioms), file=stream)
        for axiom in self.axioms:
            axiom.output(stream)

    def get_encoding_size(self):
        task_size = 0
        task_size += self.variables.get_encoding_size()
        for mutex in self.mutexes:
            task_size += mutex.get_encoding_size()
        task_size += self.goal.get_encoding_size()
        for op in self.operators:
            task_size += op.get_encoding_size()
        for axiom in self.axioms:
            task_size += axiom.get_encoding_size()
        return task_size


class SASVariables:
    def __init__(self, ranges, axiom_layers, value_names):
        self.ranges = ranges
        self.axiom_layers = axiom_layers
        self.value_names = value_names

    def validate(self):
        """Validate variables.

        All variables must have range at least 2, and derived
        variables must have range exactly 2. See comment on derived
        variables in the docstring of SASTask.validate.
        """
        assert len(self.ranges) == len(self.axiom_layers) == len(
            self.value_names)
        for (var_range, layer, var_value_names) in zip(
                self.ranges, self.axiom_layers, self.value_names):
            assert var_range == len(var_value_names)
            assert var_range >= 2
            assert layer == -1 or layer >= 0
            if layer != -1:
                assert var_range == 2

    def validate_fact(self, fact):
        """Assert that fact is a valid (var, value) pair."""
        var, value = fact
        assert 0 <= var < len(self.ranges)
        assert 0 <= value < self.ranges[var]

    def validate_condition(self, condition):
        """Assert that the condition (list of facts) is sorted, mentions each
        variable at most once, and only consists of valid facts."""
        last_var = -1
        for (var, value) in condition:
            self.validate_fact((var, value))
            assert var > last_var
            last_var = var

    def dump(self):
        for var, (rang, axiom_layer) in enumerate(
                zip(self.ranges, self.axiom_layers)):
            if axiom_layer != -1:
                axiom_str = " [axiom layer %d]" % axiom_layer
            else:
                axiom_str = ""
            print("v%d in {%s}%s" % (var, list(range(rang)), axiom_str))

    def output(self, stream):
        print(len(self.ranges), file=stream)
        for var, (rang, axiom_layer, values) in enumerate(zip(
                self.ranges, self.axiom_layers, self.value_names)):
            print("begin_variable", file=stream)
            print("var%d" % var, file=stream)
            print(axiom_layer, file=stream)
            print(rang, file=stream)
            assert rang == len(values), (rang, values)
            for value in values:
                print(value, file=stream)
            print("end_variable", file=stream)

    def get_encoding_size(self):
        # A variable with range k has encoding size k + 1 to also give the
        # variable itself some weight.
        return len(self.ranges) + sum(self.ranges)


class SASMutexGroup:
    def __init__(self, facts):
        self.facts = sorted(facts)

    def validate(self, variables):
        """Assert that the facts in the mutex group are sorted and unique
        and that they are all valid."""
        for fact in self.facts:
            variables.validate_fact(fact)
        assert self.facts == sorted(set(self.facts))

    def dump(self):
        for var, val in self.facts:
            print("v%d: %d" % (var, val))

    def output(self, stream):
        print("begin_mutex_group", file=stream)
        print(len(self.facts), file=stream)
        for var, val in self.facts:
            print(var, val, file=stream)
        print("end_mutex_group", file=stream)

    def get_encoding_size(self):
        return len(self.facts)


class SASInit:
    def __init__(self, values):
        self.values = values

    def validate(self, variables):
        """Validate initial state.

        Assert that the initial state contains the correct number of
        values and that all values are in range.
        """

        assert len(self.values) == len(variables.ranges)
        for fact in enumerate(self.values):
            variables.validate_fact(fact)

    def dump(self):
        for var, val in enumerate(self.values):
            print("v%d: %d" % (var, val))

    def output(self, stream):
        print("begin_state", file=stream)
        for val in self.values:
            print(val, file=stream)
        print("end_state", file=stream)


class SASGoal:
    def __init__(self, pairs):
        self.pairs = sorted(pairs)

    def validate(self, variables):
        """Assert that the goal is nonempty and a valid condition."""
        assert self.pairs
        variables.validate_condition(self.pairs)

    def dump(self):
        for var, val in self.pairs:
            print("v%d: %d" % (var, val))

    def output(self, stream):
        print("begin_goal", file=stream)
        print(len(self.pairs), file=stream)
        for var, val in self.pairs:
            print(var, val, file=stream)
        print("end_goal", file=stream)

    def get_encoding_size(self):
        return len(self.pairs)


class SASOperator:
    def __init__(self, name, prevail, pre_post, cost):
        self.name = name
        self.prevail = sorted(prevail)
        self.pre_post = self._canonical_pre_post(pre_post)
        self.cost = cost

    def _canonical_pre_post(self, pre_post):
        # Return a sorted and uniquified version of pre_post. We would
        # like to just use sorted(set(pre_post)), but this fails because
        # the effect conditions are a list and hence not hashable.
        def tuplify(entry):
            var, pre, post, cond = entry
            return var, pre, post, tuple(cond)
        def listify(entry):
            var, pre, post, cond = entry
            return var, pre, post, list(cond)
        pre_post = map(tuplify, pre_post)
        pre_post = sorted(set(pre_post))
        pre_post = list(map(listify, pre_post))
        return pre_post

    def validate(self, variables):
        """Validate the operator.

        Assert that
        1. Prevail conditions are valid conditions (i.e., sorted and
           all referring to different variables)
        2. The pre_post list is sorted by (var, pre, post, cond), and the
           same (var, pre, post, cond) 4-tuple is not repeated.
        3. Effect conditions are valid conditions and do not contain variables
           from the pre- or prevail conditions.
        4. Variables occurring in pre_post rules do not have a prevail
           condition.
        5. Preconditions in pre_post are -1 or valid facts.
        6. Effects are valid facts.
        7. Effect variables are non-derived.
        8. If a variable has multiple pre_post rules, then pre is
           identical in all these rules.
        9. There is at least one effect.
        10. Costs are non-negative integers.

        Odd things that are *not* illegal:
        - The effect in a pre_post rule may be identical to the
          precondition or to an effect condition of that effect.

        TODO/open question:
        - It is currently not very clear what the semantics of operators
          should be when effects "conflict", i.e., when multiple effects
          trigger and want to set a given variable to two different
          values. In the case where both are unconditional effects, we
          should make sure that our representation doesn't actually
          contain two such effects, but when at least one of them is
          conditional, things are not so easy.

          To make our life simpler when generating SAS+ tasks from
          PDDL tasks, it probably makes most sense to generalize the
          PDDL rule in this case: there is a value order where certain
          values "win" over others in this situation. It probably
          makes sense to say the "highest" values should win in this
          case, because that's consistent with the PDDL rules if we
          say false = 0 and true = 1, and also with our sort order of
          effects it means we get the right result if we just apply
          effects in sequence.

          But whatever we end up deciding, we need to be clear about it,
          document it and make sure that all of our code knows the rules
          and follows them.
        """

        variables.validate_condition(self.prevail)
        assert self.pre_post == self._canonical_pre_post(self.pre_post)
        prevail_vars = {var for (var, value) in self.prevail}
        pre_values = {}
        for var, pre, post, cond in self.pre_post:
            variables.validate_condition(cond)
            assert var not in prevail_vars
            if pre != -1:
                variables.validate_fact((var, pre))
            variables.validate_fact((var, post))
            assert variables.axiom_layers[var] == -1
            if var in pre_values:
                assert pre_values[var] == pre
            else:
                pre_values[var] = pre
        for var, pre, post, cond in self.pre_post:
            for cvar, cval in cond:
                assert(cvar not in pre_values or pre_values[cvar] == -1)
                assert(cvar not in prevail_vars)
        assert self.pre_post
        assert self.cost >= 0 and self.cost == int(self.cost)

    def dump(self):
        print(self.name)
        print("Prevail:")
        for var, val in self.prevail:
            print("  v%d: %d" % (var, val))
        print("Pre/Post:")
        for var, pre, post, cond in self.pre_post:
            if cond:
                cond_str = " [%s]" % ", ".join(
                    ["%d: %d" % tuple(c) for c in cond])
            else:
                cond_str = ""
            print("  v%d: %d -> %d%s" % (var, pre, post, cond_str))

    def output(self, stream):
        print("begin_operator", file=stream)
        print(self.name[1:-1], file=stream)
        print(len(self.prevail), file=stream)
        for var, val in self.prevail:
            print(var, val, file=stream)
        print(len(self.pre_post), file=stream)
        for var, pre, post, cond in self.pre_post:
            print(len(cond), end=' ', file=stream)
            for cvar, cval in cond:
                print(cvar, cval, end=' ', file=stream)
            print(var, pre, post, file=stream)
        print(self.cost, file=stream)
        print("end_operator", file=stream)

    def get_encoding_size(self):
        size = 1 + len(self.prevail)
        for var, pre, post, cond in self.pre_post:
            size += 1 + len(cond)
            if pre != -1:
                size += 1
        return size

    def get_applicability_conditions(self):
        """Return the combined applicability conditions
        (prevail conditions and preconditions) of the operator.

        Returns a sorted list of (var, value) pairs. This is
        guaranteed to contain at most one fact per variable and
        must hence be non-contradictory."""
        conditions = {}
        for var, val in self.prevail:
            assert var not in conditions
            conditions[var] = val
        for var, pre, post, cond in self.pre_post:
            if pre != -1:
                assert var not in conditions or conditions[var] == pre
                conditions[var] = pre
        return sorted(conditions.items())


class SASAxiom:
    def __init__(self, condition, effect):
        self.condition = sorted(condition)
        self.effect = effect
        assert self.effect[1] in (0, 1)

        for _, val in condition:
            assert val >= 0, condition

    def validate(self, variables, init):

        """Validate the axiom.

        Assert that the axiom condition is a valid condition, that the
        effect is a valid fact, that the effect variable is a derived
        variable, and that the layering condition is satisfied.

        See the docstring of SASTask.validate for information on the
        restriction on derived variables. The layering condition boils
        down to:

        1. Axioms always set the "non-init" value of the derived
           variable.
        2. Derived variables in the condition must have a lower of
           equal layer to derived variables appearing in the effect.
        3. Conditions with equal layer are only allowed when the
           condition uses the "non-init" value of that variable.

        TODO/bug: rule #1 is currently disabled because we currently
        have axioms that violate it. This is likely due to the
        "extended domain transition graphs" described in the Fast
        Downward paper, Section 5.1. However, we want to eventually
        changes this. See issue454. For cases where rule #1 is violated,
        "non-init" should be "init" in rule #3.
        """

        variables.validate_condition(self.condition)
        variables.validate_fact(self.effect)
        eff_var, eff_value = self.effect
        eff_layer = variables.axiom_layers[eff_var]
        assert eff_layer >= 0
        eff_init_value = init.values[eff_var]
        ## The following rule is currently commented out because of
        ## the TODO/bug mentioned in the docstring.
        # assert eff_value != eff_init_value
        for cond_var, cond_value in self.condition:
            cond_layer = variables.axiom_layers[cond_var]
            if cond_layer != -1:
                assert cond_layer <= eff_layer
                if cond_layer == eff_layer:
                    cond_init_value = init.values[cond_var]
                    ## Once the TODO/bug above is addressed, the
                    ## following four lines can be simplified because
                    ## we are guaranteed to land in the "if" branch.
                    if eff_value != eff_init_value:
                        assert cond_value != cond_init_value
                    else:
                        assert cond_value == cond_init_value

    def dump(self):
        print("Condition:")
        for var, val in self.condition:
            print("  v%d: %d" % (var, val))
        print("Effect:")
        var, val = self.effect
        print("  v%d: %d" % (var, val))

    def output(self, stream):
        print("begin_rule", file=stream)
        print(len(self.condition), file=stream)
        for var, val in self.condition:
            print(var, val, file=stream)
        var, val = self.effect
        print(var, 1 - val, val, file=stream)
        print("end_rule", file=stream)

    def get_encoding_size(self):
        return 1 + len(self.condition)



from machetli.planning.downward_lib.sas_tasks import SASTask, SASVariables, SASMutexGroup, SASInit, SASGoal, SASOperator, SASAxiom


def read_variables(sf, num_vars):
    axiom_layers = []
    ranges = []
    value_name_lists = []
    for var in range(num_vars):
        beginning_line = sf.readline().replace("\n", "")
        assert beginning_line == "begin_variable"
        sf.readline()  # skip variable name
        axiom_layers.append(int(sf.readline().replace("\n", "")))
        ranges.append(int(sf.readline().replace("\n", "")))
        value_names = []
        for name in range(ranges[-1]):
            value_names.append(sf.readline().replace("\n", ""))
        value_name_lists.append(value_names)
        ending_line = sf.readline().replace("\n", "")
        assert ending_line == "end_variable"
    return SASVariables(ranges, axiom_layers, value_name_lists)


def read_mutexes(sf, num_mutexes):
    mutexes = []
    for mutex_group in range(num_mutexes):
        beginning_line = sf.readline().replace("\n", "")
        assert beginning_line == "begin_mutex_group"
        num_facts = int(sf.readline().replace("\n", ""))
        facts = []
        for fact in range(num_facts):
            var, val = map(int, sf.readline().replace("\n", "").split(" "))
            facts.append((var, val))
        mutexes.append(SASMutexGroup(facts))
        ending_line = sf.readline().replace("\n", "")
        assert ending_line == "end_mutex_group"
    return mutexes


def read_init_state(sf, num_vars):
    init = []
    beginning_line = sf.readline().replace("\n", "")
    assert beginning_line == "begin_state"
    for var in range(num_vars):
        val = int(sf.readline().replace("\n", ""))
        init.append(val)
    ending_line = sf.readline().replace("\n", "")
    assert ending_line == "end_state"
    return SASInit(init)


def read_goal(sf):
    beginning_line = sf.readline().replace("\n", "")
    assert beginning_line == "begin_goal"
    num_pairs = int(sf.readline().replace("\n", ""))
    pairs = []
    for pair in range(num_pairs):
        var, val = map(int, sf.readline().replace("\n", "").split(" "))
        pairs.append((var, val))
    ending_line = sf.readline().replace("\n", "")
    assert ending_line == "end_goal"
    return SASGoal(pairs)


def read_operators(sf, num_operators):
    operators = []
    for op in range(num_operators):
        beginning_line = sf.readline().replace("\n", "")
        assert beginning_line == "begin_operator"
        name = "(" + sf.readline().replace("\n", "") + ")"
        num_prevail_conditions = int(sf.readline().replace("\n", ""))
        prevail_conditions = []
        for pre_cond in range(num_prevail_conditions):
            var, val = map(int, sf.readline().replace("\n", "").split(" "))
            prevail_conditions.append((var, val))
        num_effects = int(sf.readline().replace("\n", ""))
        pre_post = []
        for eff in range(num_effects):
            effect_line = list(map(int, sf.readline().replace("\n", "").split(" ")))
            num_effect_conditions = effect_line[0]
            cond = []
            for cond_num in range(1, 2 * num_effect_conditions, 2):
                var = effect_line[cond_num]
                val = effect_line[cond_num + 1]
                cond.append((var, val))
            var, pre, post = effect_line[-3:]
            pre_post.append((var, pre, post, cond))
        cost = int(sf.readline().replace("\n", ""))
        operators.append(SASOperator(name, prevail_conditions, pre_post, cost))
        ending_line = sf.readline().replace("\n", "")
        assert ending_line == "end_operator"
    return operators


def read_axioms(sf, num_axioms):
    axioms = []
    for ax in range(num_axioms):
        beginning_line = sf.readline().replace("\n", "")
        assert beginning_line == "begin_rule"
        length_body = int(sf.readline().replace("\n", ""))
        condition = []
        for fact in range(length_body):
            var, val = map(int, sf.readline().replace("\n", "").split(" "))
            condition.append((var, val))
        effect_line = list(map(int, sf.readline().replace("\n", "").split(" ")))
        var = effect_line[0]
        val = effect_line[2]
        assert 1 - val == effect_line[1]
        effect = (var, val)
        axioms.append(SASAxiom(condition, effect))
        ending_line = sf.readline().replace("\n", "")
        assert ending_line == "end_rule"
    return axioms


def sas_file_to_SASTask(sas_file) -> SASTask:
    with open(sas_file, "r") as sf:
        while True:
            # pos = sf.tell()
            line = sf.readline().replace("\n", "")
            if line == "begin_metric":
                break
        metric = bool(sf.readline().replace("\n", ""))
        sf.readline()  # skip end_metric
        # read variables
        num_vars = int(sf.readline().replace("\n", ""))
        variables = read_variables(sf, num_vars)
        # read mutexes
        num_mutexes = int(sf.readline().replace("\n", ""))
        mutexes = read_mutexes(sf, num_mutexes)
        # read init state
        init = read_init_state(sf, num_vars)
        # read goal
        goal = read_goal(sf)
        # read operators
        num_operators = int(sf.readline().replace("\n", ""))
        operators = read_operators(sf, num_operators)
        # read axioms
        num_axioms = int(sf.readline().replace("\n", ""))
        axioms = read_axioms(sf, num_axioms)

    sas_task = SASTask(variables, mutexes, init, goal, operators, axioms, metric)
    sas_task.validate()
    return sas_task


def write_SAS(sas_task, filename):
    with open(filename, "w") as file:
        sas_task.output(file)



#! /usr/bin/env python3


class Graph:
    def __init__(self, nodes):
        self.nodes = nodes
        self.neighbours = {u: set() for u in nodes}
    def connect(self, u, v):
        self.neighbours[u].add(v)
        self.neighbours[v].add(u)
    def connected_components(self):
        remaining_nodes = set(self.nodes)
        result = []
        def dfs(node):
            result[-1].append(node)
            remaining_nodes.remove(node)
            for neighbour in self.neighbours[node]:
                if neighbour in remaining_nodes:
                    dfs(neighbour)
        while remaining_nodes:
            node = next(iter(remaining_nodes))
            result.append([])
            dfs(node)
            result[-1].sort()
        return sorted(result)


def transitive_closure(pairs):
    # Warshall's algorithm.
    result = set(pairs)
    nodes = {u for (u, v) in pairs} | {v for (u, v) in pairs}
    for k in nodes:
        for i in nodes:
            for j in nodes:
                if (i, j) not in result and (i, k) in result and (k, j) in result:
                    result.add((i, j))
    return sorted(result)


if __name__ == "__main__":
    g = Graph([1, 2, 3, 4, 5, 6])
    g.connect(1, 2)
    g.connect(1, 3)
    g.connect(4, 5)
    print(g.connected_components())



from machetli.planning.downward_lib.pddl import Task, Truth
from machetli.planning.downward_lib.pddl.conditions import ConstantCondition, Atom

CLOSING_BRACKET = ")"
SIN = " "  # single indentation
DIN = "  "  # double indentation


def write_domain_header(task, df):
    df.write("define (domain {})\n".format(task.domain_name))


def write_domain_requirements(task, df):
    if len(task.requirements.requirements) != 0:
        df.write(SIN + "(:requirements")
        for req in task.requirements.requirements:
            df.write(" " + req)
        df.write(")\n")


def write_domain_types(task, df):
    if task.types:
        df.write(SIN + "(:types\n")
        types_dict = {}
        for tp in task.types:  # build dictionary of base types and types
            if tp.basetype_name is not None:
                if tp.basetype_name not in types_dict:
                    types_dict[tp.basetype_name] = [tp.name]
                else:
                    types_dict[tp.basetype_name].append(tp.name)
        for basetype in types_dict:
            df.write(SIN + DIN)
            for name in types_dict[basetype]:
                df.write(name + " ")
            df.write("- " + basetype + "\n")
        df.write(SIN + ")\n")


def write_domain_objects(task, df):
    if task.objects:  # all objects from planning task are going to be written into constants
        df.write(SIN + "(:constants\n")
        objects_dict = {}
        for obj in task.objects:  # build dictionary of object type names and object names
            if obj.type_name not in objects_dict:
                objects_dict[obj.type_name] = [obj.name]
            else:
                objects_dict[obj.type_name].append(obj.name)
        for type_name in objects_dict:
            df.write(SIN + DIN)
            for name in objects_dict[type_name]:
                df.write(name + " ")
            df.write("- " + type_name + "\n")
        df.write(SIN + ")\n")


def write_domain_predicates(task, df):
    if len(task.predicates) != 0:
        df.write(SIN + "(:predicates\n")
        for pred in task.predicates:
            if pred.name == "=":
                continue
            types_dict = {}
            for arg in pred.arguments:
                if arg.type_name not in types_dict:
                    types_dict[arg.type_name] = [arg.name]
                else:
                    types_dict[arg.type_name].append(arg.name)
            df.write(SIN + SIN + "(" + pred.name)
            for obj in types_dict:
                for name in types_dict[obj]:
                    df.write(" " + name)
                df.write(" - " + obj)
            df.write(")\n")
        df.write(SIN + ")\n")


def write_domain_functions(task, df):
    if task.functions:
        df.write(SIN + "(:functions\n")
        for function in task.functions:
            function.dump_pddl(df, DIN)
        df.write(SIN + ")\n")


def write_domain_actions(task, df):
    for action in task.actions:
        df.write(SIN + "(:action {}\n".format(action.name))

        df.write(DIN + ":parameters (")
        if action.parameters:
            for par in action.parameters:
                df.write("%s - %s " % (par.name, par.type_name))
        df.write(")\n")

        df.write(SIN + SIN + ":precondition\n")
        if not isinstance(action.precondition, Truth):
            action.precondition.dump_pddl(df, DIN)
        df.write(DIN + ":effect\n")
        df.write(DIN + "(and\n")
        for eff in action.effects:
            eff.dump_pddl(df, DIN)
        if action.cost:
            action.cost.dump_pddl(df, DIN + DIN)
        df.write(DIN + ")\n")

        df.write(SIN + ")\n")


def write_domain_axioms(task, df):
    for axiom in task.axioms:
        df.write(SIN + "(:derived ({} ".format(axiom.name))
        for par in axiom.parameters:
            df.write("%s - %s " % (par.name, par.type_name))
        df.write(")\n")
        axiom.condition.dump_pddl(df, DIN)
        df.write(SIN + ")\n")


def write_domain_PDDL(task, domain_filename):
    with open(domain_filename, "w") as df:
        df.write("\n(")
        write_domain_header(task, df)
        write_domain_requirements(task, df)
        write_domain_types(task, df)
        write_domain_objects(task, df)
        write_domain_predicates(task, df)
        write_domain_functions(task, df)
        write_domain_axioms(task, df)
        write_domain_actions(task, df)
        df.write(")\n")


def write_problem_header(task, pf):
    pf.write("define (problem {})\n".format(task.task_name))


def write_problem_domain(task, pf):
    pf.write(SIN + "(:domain {})\n".format(task.domain_name))


def write_problem_init(task, pf):
    pf.write(SIN + "(:init\n")

    for elem in task.init:
        if isinstance(elem, Atom) and elem.predicate == "=":
            continue
        elem.dump_pddl(pf, SIN + DIN)
    pf.write(SIN + ")\n")


def write_problem_goal(task, pf):
    pf.write(SIN + "(:goal\n")
    if not isinstance(task.goal, ConstantCondition):
        task.goal.dump_pddl(pf, SIN + DIN)
    pf.write("%s)\n" % SIN)


def write_problem_metric(task, pf):
    if task.use_min_cost_metric:
        pf.write("%s(:metric minimize (total-cost))\n" % SIN)


def write_problem_PDDL(task, problem_filename):
    with open(problem_filename, "w") as pf:
        pf.write("\n(")
        write_problem_header(task, pf)
        write_problem_domain(task, pf)
        write_problem_init(task, pf)
        write_problem_goal(task, pf)
        write_problem_metric(task, pf)
        pf.write(")\n")


def write_PDDL(task: Task, domain_filename: str, problem_filename: str):
    write_domain_PDDL(task, domain_filename)
    write_problem_PDDL(task, problem_filename)



import logging
import os
import pprint
import re
import subprocess
import time

from machetli import tools
from machetli.evaluator import is_evaluator_successful
from machetli.grid import slurm_tools as st
from machetli.tools import SubmissionError, TaskError, PollingError, write_state


EVAL_DIR = "eval_dir"
STATE_FILENAME = "state.pickle"
SBATCH_FILE = "slurm-array-job.sbatch"
FILESYSTEM_TIME_INTERVAL = 3
FILESYSTEM_TIME_LIMIT = 60
POLLING_TIME_INTERVAL = 15
TIME_LIMIT_FACTOR = 1.5

# Sets of slurm job state codes
DONE_STATE = {"COMPLETED"}
BUSY_STATES = {"PENDING", "RUNNING", "REQUEUED", "SUSPENDED"}


"""
When performing the search on a Slurm grid, the possibility of
failure at some point is increased due to the introduced parallelism
on multiple nodes and an I/O load over the network filesystem. When
setting *allow_nondeterministic_successor_choice* to ``False``, the
:func:`search <machetli.search.search>` function will enforce that
the search is aborted if a single task fails and no successor from
an earlier task is accepted.
"""
class Environment:
    def __init__(self, allow_nondeterministic_successor_choice=True,
                 batch_size=1, loglevel=logging.INFO):
        self.batch_size = batch_size
        self.loglevel = loglevel
        self.allow_nondeterministic_successor_choice = \
            allow_nondeterministic_successor_choice

    def submit(self, batch, batch_id, evaluator_path):
        raise NotImplementedError

    def wait_until_finished(self):
        raise NotImplementedError

    def get_improving_successor(self):
        raise NotImplementedError


class LocalEnvironment(Environment):
    def __init__(self, **kwargs):
        Environment.__init__(self, **kwargs)
        self.successor = None

    def submit(self, batch, batch_id, evaluator_path):
        assert self.successor is None

        for state in batch:
            if is_evaluator_successful(evaluator_path, state):
                self.successor = state
            break

    def wait_until_finished(self):
        pass

    def get_improving_successor(self):
        result = self.successor
        self.successor = None
        return result


class SlurmEnvironment(Environment):
    # Must be overridden in derived classes.
    DEFAULT_PARTITION = None
    DEFAULT_QOS = None
    DEFAULT_MEMORY_PER_CPU = None

    # Can be overridden in derived classes.
    DEFAULT_EXPORT = ["PATH"]
    DEFAULT_SETUP = ""
    DEFAULT_NICE = 0

    # TODO: are differences to Lab reasonable? e.g., here we have no time limit.
    def __init__(
        self,
        email=None,
        extra_options=None,
        partition=None,
        qos=None,
        memory_per_cpu=None,
        cpus_per_task=1,
        nice=None,
        export=None,
        setup=None,
        batch_size=200,
        **kwargs
    ):
        Environment.__init__(self, batch_size=batch_size, **kwargs)

        self.email = email
        self.extra_options = extra_options or "## (not used)"
        self.partition = partition or self.DEFAULT_PARTITION
        self.qos = qos or self.DEFAULT_QOS
        self.memory_per_cpu = memory_per_cpu or self.DEFAULT_MEMORY_PER_CPU
        self.cpus_per_task = cpus_per_task
        self.nice = nice or self.DEFAULT_NICE
        self.export = export or self.DEFAULT_EXPORT
        self.setup = setup or self.DEFAULT_SETUP
        self.script_path = tools.get_script_path()
        self.job = None

        script_dir = os.path.dirname(self.script_path)
        self.eval_dir = os.path.join(script_dir, EVAL_DIR)
        tools.makedirs(self.eval_dir)
        st.check_for_whitespace(self.eval_dir)
        self.sbatch_file = os.path.join(script_dir, SBATCH_FILE)
        self.wait_for_filesystem(self.eval_dir)
        self.critical = False

    def get_job_params(self):
        job_params = dict()
        job_params["logfile"] = "slurm.log"
        job_params["errfile"] = "slurm.err"
        job_params["partition"] = self.partition
        job_params["qos"] = self.qos
        job_params["memory_per_cpu"] = self.memory_per_cpu
        job_params["nice"] = self.nice
        job_params["extra_options"] = self.extra_options
        job_params["environment_setup"] = self.setup
        job_params["mailtype"] = "NONE"
        job_params["mailuser"] = ""
        job_params["soft_memory_limit"] = int(
            0.98 * self.cpus_per_task * self._get_memory_in_kb(
                self.memory_per_cpu))
        job_params["python"] = tools.get_python_executable()
        job_params["script_path"] = self.script_path
        return job_params

    def submit(self, batch, batch_id, evaluator_path):
        assert not self.job
        try:
            self.job = self.submit_array_job(batch, batch_id, evaluator_path)
        except SubmissionError as se:
            if self.allow_nondeterministic_successor_choice:
                se.warn_abort()
                # TODO: this means job is undefined, so we should also abort.
            else:
                se.warn()
                self.job = None
                raise se

    def wait_until_finished(self):
        assert self.job
        try:
            self.poll_job()
        except TaskError as te:
            if self.allow_nondeterministic_successor_choice:
                te.remove_critical_tasks(self.job)
                if not self.job["tasks"]:
                    # TODO: this is just a hack to replace the "continue"
                    #  that occurred here in the original grid search.
                    self.critical = True
                    return
            else:
                te.remove_tasks_after_first_critical(self.job)
                if not self.job["tasks"]:
                    self.job = None
                    raise te
        except PollingError as pe:
            pe.warn_abort()
            raise pe

    def get_improving_successor(self):
        assert self.job
        if self.critical:
            self.critical = False
            self.job = None
            return None

        successor = None
        for task in self.job["tasks"]:
            result_file = os.path.join(task["dir"], "exit_code")
            if self.wait_for_filesystem(result_file):
                if st.parse_exit_code(result_file) == 0:
                    successor = task["state"]
                    break
            else:
                if self.allow_nondeterministic_successor_choice:
                    logging.warning(
                        f"Result file {result_file} does not exist. "
                        "Continuing with next task.")
                    continue
                else:
                    logging.warning("Aborting search because evaluation "
                                    f"in {task['dir']} failed.")
                    # TODO: raise an error that can be handled by the caller.
                    return None
        self.job = None
        return successor

    def wait_for_filesystem(self, *paths):
        attempts = int(FILESYSTEM_TIME_LIMIT / FILESYSTEM_TIME_INTERVAL)
        for _ in range(attempts):
            time.sleep(FILESYSTEM_TIME_INTERVAL)
            paths = [path for path in paths if not os.path.exists(path)]
            if not paths:
                return True
        return False  # At least one path from paths does not exist

    def build_batch_directories(self, batch, batch_num):
        batch_dir_path = os.path.join(
            self.eval_dir, f"batch_{batch_num:03}")
        run_dirs = []
        for rank, state in enumerate(batch):
            run_dir_name = f"{rank:03}"
            run_dir_path = os.path.join(batch_dir_path, run_dir_name)
            tools.makedirs(run_dir_path)
            state_file_path = os.path.join(run_dir_path, STATE_FILENAME)
            write_state(state, state_file_path)
            run_dirs.append(run_dir_path)
        # Give the NFS time to write the paths
        if not self.wait_for_filesystem(*run_dirs):
            logging.critical(
                f"One of the following paths is missing:\n"
                f"{pprint.pformat(run_dirs)}"
            )
        return run_dirs

    def write_sbatch_file(self, **kwargs):
        dictionary = self.get_job_params()
        dictionary.update(kwargs)
        logging.debug(
            f"Dictionary before filling:\n{pprint.pformat(dictionary)}")
        filled_text = st.fill_template(**dictionary)
        with open(self.sbatch_file, "w") as f:
            f.write(filled_text)
        # TODO: Implement check whether file was updated

    def submit_array_job(self, batch, batch_num, evaluator_path):
        """
        Writes pickled version of each state in *batch* to its own file.
        Then, submits a slurm array job which will evaluate each state
        in parallel. Returns the array job ID of the submitted array job.
        """
        run_dirs = self.build_batch_directories(batch, batch_num)
        batch_name = f"batch_{batch_num:03}"
        self.write_sbatch_file(run_dirs=" ".join(run_dirs), name=batch_name,
                               num_tasks=len(batch)-1, evaluator_path=evaluator_path)
        submission_command = ["sbatch", "--export",
                              ",".join(self.export), self.sbatch_file]
        try:
            output = subprocess.check_output(submission_command).decode()
        except subprocess.CalledProcessError as cpe:
            raise SubmissionError(cpe)
        match = re.match(r"Submitted batch job (\d*)", output)
        if not match:
            logging.critical(
                "Something went wrong, no job ID printed after job submission.")
        else:
            logging.info(match.group(0))
        job_id = match.group(1)
        job = {"id": job_id,
               "tasks": [{"state": s, "dir": d} for s, d in
                         zip(batch, run_dirs)]}
        return job

    def poll_job(self):
        job_id = self.job["id"]
        while True:
            time.sleep(POLLING_TIME_INTERVAL)
            try:
                output = subprocess.check_output(
                    ["sacct", "-j", str(job_id), "--format=jobid,state",
                     "--noheader", "--allocations"]).decode()
                task_states = self._build_task_state_dict(output)
                done = []
                busy = []
                critical = []
                logging.debug(f"Task states:\n{pprint.pformat(task_states)}")
                for task_id, task_state in task_states.items():
                    if task_state in DONE_STATE:
                        done.append(task_id)
                    elif task_state in BUSY_STATES:
                        busy.append(task_id)
                    else:
                        critical.append(task_id)
                if busy:
                    logging.info(
                        f"{len(busy)} task"
                        f"{'s are' if len(busy) > 1 else ' is'} still busy.")
                    continue
                if critical:
                    critical_tasks = [
                        task for task in task_states if task in critical]
                    raise TaskError(critical_tasks)
                else:
                    logging.info("Batch completed.")
                    return
            except subprocess.CalledProcessError:
                raise PollingError(job_id)

    def _build_task_state_dict(self, sacct_output):
        unclean_job_state_list = sacct_output.strip("\n").split("\n")
        stripped_job_state_list = [pair.strip(
            "+ ") for pair in unclean_job_state_list]
        return {k: v for k, v in
                (pair.split() for pair in stripped_job_state_list)}

    @staticmethod
    # This function is copied from lab.environment.SlurmEnvironment
    # (<https://lab.readthedocs.org>).
    def _get_memory_in_kb(limit):
        match = re.match(r"^(\d+)(k|m|g)?$", limit, flags=re.I)
        if not match:
            logging.critical(f"malformed memory_per_cpu parameter: {limit}")
        memory = int(match.group(1))
        suffix = match.group(2)
        if suffix is not None:
            suffix = suffix.lower()
        if suffix == "k":
            pass
        elif suffix is None or suffix == "m":
            memory *= 1024
        elif suffix == "g":
            memory *= 1024 * 1024
        return memory


class BaselSlurmEnvironment(SlurmEnvironment):
    """Environment for Basel's AI group."""
    DEFAULT_PARTITION = "infai_1"
    DEFAULT_QOS = "normal"
    DEFAULT_MEMORY_PER_CPU = "3872M"
    MAX_MEM_INFAI_BASEL = {"infai_1": "3872M", "infai_2": "6354M"}
    DEFAULT_NICE = 5000

    def __init__(self, **kwargs):
        SlurmEnvironment.__init__(self, **kwargs)

        # Abort if mem_per_cpu too high for Basel partitions
        if self.partition in {"infai_1", "infai_2"}:
            mem_per_cpu_in_kb = self._get_memory_in_kb( self.memory_per_cpu)
            max_mem_per_cpu_in_kb = self._get_memory_in_kb(
                self.MAX_MEM_INFAI_BASEL[self.partition])
            if mem_per_cpu_in_kb > max_mem_per_cpu_in_kb:
                logging.critical(
                    f"Memory limit {self.memory_per_cpu} surpassing the "
                    f"maximum amount allowed for partition {self.partition}: "
                    f"{self.MAX_MEM_INFAI_BASEL[self.partition]}."
                )




import copy
import itertools
import random

from machetli.planning import pddl_visitors
from machetli.planning.downward_lib import timers
from machetli.planning.downward_lib.sas_tasks import SASTask, SASMutexGroup, SASInit, SASGoal, SASOperator, SASAxiom
from machetli.successors import SuccessorGenerator


class RemoveActions(SuccessorGenerator):
    """Successor generator that removes 
    randomly selected actions from the PDDL task in a state.
    """
    def get_successors(self, state):
        """Yield modified versions of *state* of which in each
        one a different action is removed from the PDDL task
        stored in ``state["pddl_task"]``.
        """
        task = state["pddl_task"]
        action_names = [action.name for action in task.actions]
        random.Random().shuffle(action_names)
        for name in action_names:
            child_state = copy.deepcopy(state)
            pre_child_task = child_state["pddl_task"]
            with timers.timing("Obtaining successor"):
                child_task = pre_child_task.accept(
                    pddl_visitors.TaskElementEraseActionVisitor(name))
            child_state["pddl_task"] = child_task
            yield child_state


class ReplaceAtomsWithTruth(SuccessorGenerator):
    """Successor generator that removes 
    randomly selected atoms from the PDDL task in a state.
    This is accomplished by scanning the entire task for the
    atom to be removed, instantiating each instance of this atom
    with the *truth* value and then simplifying all logical expressions.
    """
    def get_successors(self, state):
        """Yield modified versions of *state* of which in each
        one a different atom is removed from the PDDL task
        stored in ``state["pddl_task"]``. 
        """
        task = state["pddl_task"]
        predicate_names = [predicate.name for predicate in task.predicates if
                           not (predicate.name == "dummy_axiom_trigger" or predicate.name == "=")]
        random.Random().shuffle(predicate_names)
        for name in predicate_names:
            child_state = copy.deepcopy(state)
            pre_child_task = child_state["pddl_task"]
            with timers.timing("Obtaining successor"):
                child_task = pre_child_task.accept(
                    pddl_visitors.TaskElementErasePredicateTrueAtomVisitor(name))
            child_state["pddl_task"] = child_task
            yield child_state


class ReplaceAtomsWithFalsity(SuccessorGenerator):
    """Successor generator that removes 
    randomly selected atoms from the PDDL task in a state.
    The same mechanism is used as in :class:`ReplaceAtomsWithTruth <machetli.planning.generators.ReplaceAtomsWithTruth>`,
    but replacing atoms with *falsity* instead.
    """
    def get_successors(self, state):
        """Yield modified versions of *state* of which in each
        one a different atom is removed from the PDDL task
        stored in ``state["pddl_task"]``. 
        """
        task = state["pddl_task"]
        predicate_names = [predicate.name for predicate in task.predicates if
                           not (predicate.name == "dummy_axiom_trigger" or predicate.name == "=")]
        random.Random().shuffle(predicate_names)
        for name in predicate_names:
            child_state = copy.deepcopy(state)
            pre_child_task = child_state["pddl_task"]
            with timers.timing("Obtaining successor"):
                child_task = pre_child_task.accept(
                    pddl_visitors.TaskElementErasePredicateFalseAtomVisitor(name))
            child_state["pddl_task"] = child_task
            yield child_state


class ReplaceLiteralsWithTruth(SuccessorGenerator):
    def get_successors(self, state):
        task = state["pddl_task"]
        predicate_names = [predicate.name for predicate in task.predicates if
                           not (predicate.name == "dummy_axiom_trigger" or predicate.name == "=")]
        random.Random().shuffle(predicate_names)
        for name in predicate_names:
            child_state = copy.deepcopy(state)
            pre_child_task = child_state["pddl_task"]
            with timers.timing("Obtaining successor"):
                child_task = pre_child_task.accept(
                    pddl_visitors.TaskElementErasePredicateTrueLiteralVisitor(name))
            child_state["pddl_task"] = child_task
            yield child_state


class RemoveObjects(SuccessorGenerator):
    def get_successors(self, state):
        task = state["pddl_task"]
        object_names = [obj.name for obj in task.objects]
        random.Random().shuffle(object_names)
        for name in object_names:
            child_state = copy.deepcopy(state)
            pre_child_task = child_state["pddl_task"]
            with timers.timing("Obtaining successor"):
                child_task = pre_child_task.accept(
                    pddl_visitors.TaskElementEraseObjectVisitor(name))
            child_state["pddl_task"] = child_task
            yield child_state


class RemoveSASOperators(SuccessorGenerator):
    def get_successors(self, state):
        task = state["sas_task"]
        operator_names = [op.name for op in task.operators]
        random.Random().shuffle(operator_names)
        for name in operator_names:
            child_state = copy.deepcopy(state)
            pre_child_task = child_state["sas_task"]
            with timers.timing("Obtaining successor"):
                child_task = self.transform(pre_child_task, name)
            child_state["sas_task"] = child_task
            yield child_state

    def transform(self, task, op_name):
        new_operators = [op for op in task.operators if not op.name == op_name]

        return SASTask(task.variables, task.mutexes, task.init, task.goal, new_operators,
                       task.axioms, task.metric)


class RemoveSASVariables(SuccessorGenerator):
    def get_successors(self, state):
        task = state["sas_task"]
        variables = [var for var in range(len(task.variables.axiom_layers))]
        random.Random().shuffle(variables)
        for var in variables:
            child_state = copy.deepcopy(state)
            pre_child_task = child_state["sas_task"]
            with timers.timing("Obtaining successor"):
                child_task = self.transform(pre_child_task, var)
            child_state["sas_task"] = child_task
            yield child_state

    def transform(self, task, var):
        # remove var attributes from variables object
        new_variables = task.variables
        del new_variables.axiom_layers[var]
        del new_variables.ranges[var]
        del new_variables.value_names[var]
        # remove var from from mutex groups
        new_mutexes = []
        for group in task.mutexes:
            new_facts = []
            for fact in group.facts:
                if fact[0] == var:
                    continue
                if fact[0] > var:
                    variable_index, value = fact
                    variable_index = variable_index - 1  # decrement variable indices above var
                    fact = (variable_index, value)
                new_facts.append(fact)
            new_mutexes.append(SASMutexGroup(new_facts))
        # remove var from init
        new_init = SASInit(task.init.values)
        del new_init.values[var]
        # remove var from goal pairs and decrement higher indices than var
        new_goal_pairs = []
        for pair in task.goal.pairs:
            if pair[0] == var:
                continue
            if pair[0] > var:
                variable_index, value = pair
                variable_index = variable_index - 1  # decrement variable indices above var
                pair = (variable_index, value)
            new_goal_pairs.append(pair)
        new_goal = SASGoal(new_goal_pairs)
        # remove var from operators
        new_operators = []
        for op in task.operators:
            new_prevail = []
            for pre in op.prevail:
                if pre[0] == var:
                    continue
                if pre[0] > var:
                    variable_index, value = pre
                    variable_index = variable_index - 1  # decrement variable indices above var
                    pre = (variable_index, value)
                new_prevail.append(pre)
            new_effects = []
            for eff in op.pre_post:
                v, pre, post, cond = eff
                if v == var:
                    continue
                if v > var:
                    v = v - 1  # decrement variable indices above var
                new_cond = []
                for precondition in cond:
                    if precondition[0] == var:
                        continue
                    if precondition[0] > var:
                        variable_index, value = precondition
                        variable_index = variable_index - 1  # decrement variable indices above var
                        precondition = (variable_index, value)
                    new_cond.append(precondition)
                new_effects.append((v, pre, post, new_cond))
            if not new_effects:
                continue
            new_operators.append(SASOperator(op.name, new_prevail, new_effects, op.cost))
        # remove var from condition and effect of axioms
        new_axioms = []
        for ax in task.axioms:
            if ax.effect[0] == var:
                continue
            if ax.effect[0] > var:
                variable_index, value = ax.effect
                variable_index = variable_index - 1  # decrement variable indices above var
                ax.effect = (variable_index, value)
            new_condition = []
            for cond in ax.condition:
                if cond[0] == var:
                    continue
                if cond[0] > var:
                    variable_index, value = cond
                    variable_index = variable_index - 1  # decrement variable indices above var
                    cond = (variable_index, value)
                new_condition.append(cond)
            # axiom condition may also be empty
            new_axioms.append(SASAxiom(new_condition, ax.effect))

        return SASTask(new_variables, new_mutexes, new_init, new_goal, new_operators, new_axioms, task.metric)


class RemoveSASEffect(SuccessorGenerator):
    def get_successors(self, state):
        task = state["sas_task"]
        num_ops = len(task.operators)
        for op in random.sample(range(num_ops), num_ops):
            num_eff = len(task.operators[op].pre_post)
            for effect in random.sample(range(num_eff), num_eff):
                child_state = copy.deepcopy(state)
                with timers.timing("Obtaining successor"):
                    del child_state["sas_task"].operators[op].pre_post[effect]
                    yield child_state


class SetUnspecifiedSASPrevailCondition(SuccessorGenerator):
    def get_successors(self, state):
        task = state["sas_task"]
        num_ops = len(task.operators)
        for op in random.sample(range(num_ops), num_ops):
            num_eff = len(task.operators[op].pre_post)
            for effect in random.sample(range(num_eff), num_eff):
                var, pre, post, cond = task.operators[op].pre_post[effect]
                if pre == -1:
                    num_val = task.variables.ranges[var]
                    for val in random.sample(range(num_val), num_val):
                        child_state = copy.deepcopy(state)
                        with timers.timing("Obtaining successor"):
                            child_state["sas_task"].operators[op].pre_post[
                                effect] = (var, val, post, cond)
                            yield child_state


class MergeSASOperators(SuccessorGenerator):
    def get_successors(self, state):
        task = state["sas_task"]
        for op1, op2 in itertools.permutations(task.operators, 2):
            child_state = copy.deepcopy(state)
            with timers.timing("Obtaining successor"):
                child_task = self.transform(child_state["sas_task"], op1, op2)
            if child_task:
                child_state["sas_task"] = child_task
                yield child_state

    def transform(self, task, op1, op2):
        def combined_pre_post(op):
            combined_pre, combined_post = {}, {}
            for var, value in op.prevail:
                combined_pre[var] = value
                combined_post[var] = value
            for var, pre, post, cond in op.pre_post:
                if cond:
                    raise NotImplementedError("Conditional effects not yet supported.")
                if pre != -1:
                    combined_pre[var] = pre
                combined_post[var] = post
            return combined_pre, combined_post

        pre1, post1 = combined_pre_post(op1)
        pre2, post2 = combined_pre_post(op2)

        # Check that op2 is applicable and update preconditions
        merged_pre = pre1
        for var, pre in pre2.items():
            if var not in post1:
                merged_pre[var] = pre
            elif post1[var] != pre:
                # Operators not compatible
                return None

        # update effects
        merged_post = post1
        merged_post.update(post2)

        merged_prevail = []
        merged_pre_post = []
        for var, post in merged_post.items():
            pre = merged_pre.get(var, -1)
            if pre == post:
                merged_prevail.append((var, pre))
            else:
                merged_pre_post.append((var, pre, post, []))

        merged_name = op1.name + " and then " + op2.name
        merged_cost = op1.cost + op2.cost
        merged_op = SASOperator(merged_name, merged_prevail, merged_pre_post, merged_cost)

        new_operators = [op for op in task.operators if op.name not in [op1.name, op2.name]] + [merged_op]

        return SASTask(task.variables, task.mutexes, task.init, task.goal, new_operators,
                       task.axioms, task.metric)





from collections import Generator
import copy
import random

from machetli.planning.downward_lib import timers
from machetli.planning.downward_lib.sas_tasks import SASTask, SASMutexGroup, SASGoal, SASOperator, SASInit, SASAxiom

SEED = 42


class SASTransformer:
    """Interface for SAS+ transformer classes."""

    def get_successors(self, sas_task):
        """Return task generator representing successors of input task."""
        raise NotImplementedError("This method should be implemented.")


class SASOperatorEraser(SASTransformer):
    """Performs SAS+ task transformations by erasing SASOperators"""

    def get_successors(self, sas_task: SASTask) -> Generator:
        """Creates and returns generator of successors of given SAS+ task by deleting SASOperators."""
        operator_names = [operator.name for operator in sas_task.operators]
        random.Random(SEED).shuffle(operator_names)
        for name in operator_names:
            pre_child = copy.deepcopy(sas_task)
            with timers.timing("Obtaining successor"):
                child = self.transform(pre_child, name)
            yield child, name

    def transform(self, sas_task, op_name):
        """Transforms sas_task by deleting the operator called op_name."""
        new_operators = [op for op in sas_task.operators if not op.name == op_name]

        return SASTask(sas_task.variables, sas_task.mutexes, sas_task.init, sas_task.goal, new_operators,
                       sas_task.axioms, sas_task.metric)


class SASVariableEraser(SASTransformer):
    """Performs SAS+ task transformations by erasing SASVariables."""

    def get_successors(self, sas_task) -> Generator:
        """Creates and returns generator of successors of given SAS+ task by deleting SASVariables."""
        variables = [var for var in range(len(sas_task.variables.axiom_layers))]
        random.Random(SEED).shuffle(variables)
        for var in variables:
            pre_child = copy.deepcopy(sas_task)
            with timers.timing("Obtaining successor"):
                child = self.transform(pre_child, var)
            yield child, var

    def transform(self, sas_task, var):
        """Transforms sas_task by deleting variable with index var."""
        # remove var attributes from variables object
        new_variables = sas_task.variables
        del new_variables.axiom_layers[var]
        del new_variables.ranges[var]
        del new_variables.value_names[var]
        # remove var from from mutex groups
        new_mutexes = []
        for group in sas_task.mutexes:
            new_facts = []
            for fact in group.facts:
                if fact[0] == var:
                    continue
                if fact[0] > var:
                    variable_index, value = fact
                    variable_index = variable_index - 1  # decrement variable indices above var
                    fact = (variable_index, value)
                new_facts.append(fact)
            new_mutexes.append(SASMutexGroup(new_facts))
        # remove var from init
        new_init = SASInit(sas_task.init.values)
        del new_init.values[var]
        # remove var from goal pairs and decrement higher indices than var
        new_goal_pairs = []
        for pair in sas_task.goal.pairs:
            if pair[0] == var:
                continue
            if pair[0] > var:
                variable_index, value = pair
                variable_index = variable_index - 1  # decrement variable indices above var
                pair = (variable_index, value)
            new_goal_pairs.append(pair)
        new_goal = SASGoal(new_goal_pairs)
        # remove var from operators
        new_operators = []
        for op in sas_task.operators:
            new_prevail = []
            for pre in op.prevail:
                if pre[0] == var:
                    continue
                if pre[0] > var:
                    variable_index, value = pre
                    variable_index = variable_index - 1  # decrement variable indices above var
                    pre = (variable_index, value)
                new_prevail.append(pre)
            new_effects = []
            for eff in op.pre_post:
                v, pre, post, cond = eff
                if v == var:
                    continue
                if v > var:
                    v = v - 1  # decrement variable indices above var
                new_cond = []
                for precondition in cond:
                    if precondition[0] == var:
                        continue
                    if precondition[0] > var:
                        variable_index, value = precondition
                        variable_index = variable_index - 1  # decrement variable indices above var
                        precondition = (variable_index, value)
                    new_cond.append(precondition)
                new_effects.append((v, pre, post, new_cond))
            if not new_effects:
                continue
            new_operators.append(SASOperator(op.name, new_prevail, new_effects, op.cost))
        # remove var from condition and effect of axioms
        new_axioms = []
        for ax in sas_task.axioms:
            if ax.effect[0] == var:
                continue
            if ax.effect[0] > var:
                variable_index, value = ax.effect
                variable_index = variable_index - 1  # decrement variable indices above var
                ax.effect = (variable_index, value)
            new_condition = []
            for cond in ax.condition:
                if cond[0] == var:
                    continue
                if cond[0] > var:
                    variable_index, value = cond
                    variable_index = variable_index - 1  # decrement variable indices above var
                    cond = (variable_index, value)
                new_condition.append(cond)
            # axiom condition may also be empty
            new_axioms.append(SASAxiom(new_condition, ax.effect))

        return SASTask(new_variables, new_mutexes, new_init, new_goal, new_operators, new_axioms, sas_task.metric)



import logging
import os
import pkgutil
import re
import subprocess

from machetli import tools

TEMPLATE_FILE = "slurm-array-job.template"


def parse_exit_code(result_file):
    with open(result_file, "r") as rf:
        exitcode = int(rf.read())
    return exitcode


def fill_template(**parameters):
    template = tools.get_string(pkgutil.get_data(
        "machetli", os.path.join("grid", TEMPLATE_FILE)))
    return template.format(**parameters)


def launch_email_job(environment):
    if environment.email:
        try:
            subprocess.run(["sbatch",
                            "--job-name='Search terminated'",
                            "--mail-type=BEGIN",
                            f"--mail-user={environment.email}"],
                           input=b"#! /bin/bash\n")
        except:
            logging.warning(
                "Something went wrong while trying to send the "
                "notification email.")
    return


def check_for_whitespace(path):
    assert not re.search(r"\s+", path), \
        "The script path must not contain any whitespace characters."



class SuccessorGenerator():
    """Base class for all successor generators.
    """
    def get_successors(self, state):
        """Yield successors of *state*.
        """
        raise NotImplementedError

class ChainingSuccessorGenerator(SuccessorGenerator):
    def __init__(self, nested_generators):
        self.nested_generators = nested_generators
    
    def get_successors(self, state):
        for g in self.nested_generators:
            for s in g.get_successors(state):
                yield s

def make_single_successor_generator(generators):
    if generators is None:
        return ChainingSuccessorGenerator([])
    elif isinstance(generators, list):
        return ChainingSuccessorGenerator(generators[:])
    elif isinstance(generators, (tuple, set)):
        return ChainingSuccessorGenerator(list(generators))
    else:
        return generators



import tempfile
import contextlib
import os

from machetli.planning.sas_reader import sas_file_to_SASTask
from machetli.planning.pddl_writer import write_PDDL
from machetli.planning.downward_lib import pddl_parser

PDDL_TASK = "pddl_task"
SAS_TASK = "sas_task"


def parse_pddl_task(dom_filename, prob_filename):
    """Parse the PDDL task defined in PDDL files *dom_filename* (PDDL domain)
    and *prob_filename* (PDDL problem) an return an instance of the parsed PDDL task.

    The returned task object is an instance of the ``Task`` class used internally
    in `Fast Downward <https://www.fast-downward.org>`_.

    """
    return pddl_parser.open(domain_filename=dom_filename,
                            task_filename=prob_filename)


def parse_sas_task(task_filename):
    """Parse the SAS\ :sup:`+` task defined in the SAS\ :sup:`+` file
    *task_filename* and return an instance of the parsed SAS\ :sup:`+` task.

    The returned task object is an instance of the ``SASTask`` class used internally
    in `Fast Downward <https://www.fast-downward.org>`_.
    """
    return sas_file_to_SASTask(task_filename)

@contextlib.contextmanager
def generated_sas_file(state):
    """Context manager that generates a temporary SAS\ :sup:`+` file
    containing the task stored under the ``"sas_task"`` key in the *state*
    dictionary. After the context is left, the generated file is deleted.

    Example:

    >>> with generated_sas_file(state) as sas_filename:
    ...     cmd = ["fast-downward.py", f"{sas_filename}", "--search", "astar(lmcut())"]
    ...
    """
    f = tempfile.NamedTemporaryFile(mode="w+t", suffix=".sas", delete=False)
    state[SAS_TASK].output(f)
    f.close()
    yield f.name
    os.remove(f.name)

@contextlib.contextmanager
def generated_pddl_files(state):
    """Context manager that generates temporary PDDL files
    containing the task stored under the ``"pddl_task"`` key in the *state*
    dictionary. After the context is left, the generated files are deleted.

    Example:

    >>> with generated_pddl_files(state) as (domain_filename, problem_filename):
    ...     cmd = ["fast-downward.py", f"{domain_filename}", f"{problem_filename}", "--search", "astar(lmcut())"]
    ...
    """
    domain_f = tempfile.NamedTemporaryFile(
        mode="w+t", suffix=".pddl", delete=False)
    domain_f.close()
    problem_f = tempfile.NamedTemporaryFile(
        mode="w+t", suffix=".pddl", delete=False)
    problem_f.close()
    write_PDDL(state[PDDL_TASK], domain_filename=domain_f.name,
               problem_filename=problem_f.name)
    yield (domain_f.name, problem_f.name)
    os.remove(domain_f.name)
    os.remove(problem_f.name)



from .pddl_types import Type
from .pddl_types import TypedObject

from .tasks import Task
from .tasks import Requirements
from .task_element import TaskElement

from .predicates import Predicate

from .functions import Function

from .actions import Action
from .actions import PropositionalAction

from .axioms import Axiom
from .axioms import PropositionalAxiom

from .conditions import Literal
from .conditions import Atom
from .conditions import NegatedAtom
from .conditions import Falsity
from .conditions import Truth
from .conditions import Conjunction
from .conditions import Disjunction
from .conditions import UniversalCondition
from .conditions import ExistentialCondition

from .effects import ConditionalEffect
from .effects import ConjunctiveEffect
from .effects import CostEffect
from .effects import Effect
from .effects import SimpleEffect
from .effects import UniversalEffect

from .f_expression import Assign
from .f_expression import Increase
from .f_expression import NumericConstant
from .f_expression import PrimitiveNumericExpression



from itertools import islice
import logging

from machetli.grid.environments import LocalEnvironment
from machetli.successors import make_single_successor_generator
from machetli.tools import SubmissionError, TaskError, PollingError, configure_logging



def search(initial_state, successor_generator, evaluator_path, environment=None):
    """Start a machetli search and return the resulting state.

    The search is started from *initial_state*, which is a dictionary
    describing the initial conditions of what you want to minimize.
    
    *successor_generator* is a single :class:`SuccessorGenerator
    <machetli.planning.generators.SuccessorGenerator>` or a list of
    SuccessorGenerators. If a list [s1, ..., sn] is given, the search
    first tries all successors from s1, then from s2, and so on.

    *evaluator_path* is the path to a Python file that contains a function
    *evaluate(state)* that is used to check if the behaviour that the search
    is analyzing is still present in the state. The function has to take a single
    parameter that is the state of the search as a dictionary like the one given
    to *search* as the *initial_state* parameter. It has to return True if the
    behaviour is present and False otherwise.

    *environment* determines whether the search should be done on a local
    machine or on a Slurm computing grid. Use
    :class:`machetli.grid.environments.LocalEnvironment` or an implementation
    of :class:`machetli.grid.environments.SlurmEnvironment`. If no environment
    is specified, a local environment will be used.

    Example usage:

TODO: update line numbers and filename
    .. literalinclude:: ../examples/issue335_PDDL/local_test.py
        :language: python
        :caption:
        :lines: 18-30
        :emphasize-lines: 9-12
    """
    if environment is None:
        environment = LocalEnvironment()
    configure_logging(environment.loglevel)
    successor_generator = make_single_successor_generator(successor_generator)

    current_state = initial_state
    batch_size = environment.batch_size
    batch_num = 0

    successors = successor_generator.get_successors(current_state)
    batch = list(islice(successors, batch_size))
    while batch:
        try:
            environment.submit(batch, batch_num, evaluator_path)
            environment.wait_until_finished()
            successor = environment.get_improving_successor()
        except (SubmissionError, TaskError, PollingError):
            # FIXME: this is not proper error handling yet.
            successor = None

        if successor:
            logging.info("Found successor")
            current_state = successor
            successors = successor_generator.get_successors(current_state)
        batch_num += 1
        batch = list(islice(successors, batch_size))

    return current_state



from . import conditions
from machetli.planning.downward_lib.pddl.task_element import TaskElement


class Axiom(TaskElement):
    def accept(self, visitor):
        return visitor.visit_axiom(self)

    def __init__(self, name, parameters, num_external_parameters, condition):
        # For an explanation of num_external_parameters, see the
        # related Action class. Note that num_external_parameters
        # always equals the arity of the derived predicate.
        assert 0 <= num_external_parameters <= len(parameters)
        self.name = name
        self.parameters = parameters
        self.num_external_parameters = num_external_parameters
        self.condition = condition
        self.uniquify_variables()

    def dump(self):
        args = map(str, self.parameters[:self.num_external_parameters])
        print("Axiom %s(%s)" % (self.name, ", ".join(args)))
        self.condition.dump()

    def uniquify_variables(self):
        self.type_map = {par.name: par.type_name for par in self.parameters}
        self.condition = self.condition.uniquify_variables(self.type_map)

    def instantiate(self, var_mapping, init_facts, fluent_facts):
        # The comments for Action.instantiate apply accordingly.
        arg_list = [self.name] + [
            var_mapping[par.name]
            for par in self.parameters[:self.num_external_parameters]]
        name = "(%s)" % " ".join(arg_list)

        condition = []
        try:
            self.condition.instantiate(var_mapping, init_facts, fluent_facts, condition)
        except conditions.Impossible:
            return None

        effect_args = [var_mapping.get(arg.name, arg.name)
                       for arg in self.parameters[:self.num_external_parameters]]
        effect = conditions.Atom(self.name, effect_args)
        return PropositionalAxiom(name, condition, effect)


class PropositionalAxiom:
    def __init__(self, name, condition, effect):
        self.name = name
        self.condition = condition
        self.effect = effect

    def clone(self):
        return PropositionalAxiom(self.name, list(self.condition), self.effect)

    def dump(self):
        if self.effect.negated:
            print("not", end=' ')
        print(self.name)
        for fact in self.condition:
            print("PRE: %s" % fact)
        print("EFF: %s" % self.effect)

    @property
    def key(self):
        return (self.name, self.condition, self.effect)

    def __lt__(self, other):
        return self.key < other.key

    def __le__(self, other):
        return self.key <= other.key

    def __eq__(self, other):
        return self.key == other.key

    def __repr__(self):
        return '<PropositionalAxiom %s %s -> %s>' % (
            self.name, self.condition, self.effect)



import io
from contextlib import redirect_stdout

from machetli.planning.downward_lib.pddl import Task, TypedObject, Predicate, Action, Axiom, Function, Truth, Conjunction, \
    Disjunction, Falsity, \
    UniversalCondition, ExistentialCondition, Atom, NegatedAtom, Effect, Assign
from machetli.planning.downward_lib.pddl.conditions import ConstantCondition


class TaskElementVisitor:
    """Interface for visitor classes to visit PDDL task elements."""

    def visit_task(self, task) -> Task:
        raise NotImplementedError

    def visit_object(self, obj) -> TypedObject:
        raise NotImplementedError

    def visit_predicate(self, predicate) -> Predicate:
        raise NotImplementedError

    def visit_function(self, function) -> Function:
        raise NotImplementedError

    def visit_condition(self, condition):
        if isinstance(condition, Falsity):
            return self.visit_condition_falsity(condition)
        elif isinstance(condition, Truth):
            return self.visit_condition_truth(condition)
        elif isinstance(condition, Conjunction):
            return self.visit_condition_conjunction(condition)
        elif isinstance(condition, Disjunction):
            return self.visit_condition_disjunction(condition)
        elif isinstance(condition, UniversalCondition):
            return self.visit_condition_universal(condition)
        elif isinstance(condition, ExistentialCondition):
            return self.visit_condition_existential(condition)
        elif isinstance(condition, Atom):
            return self.visit_condition_atom(condition)
        elif isinstance(condition, NegatedAtom):
            return self.visit_condition_negated_atom(condition)
        else:
            raise NotImplementedError(
                "No visiting function implemented for this type of condition.")

    def visit_condition_falsity(self, falsity) -> Falsity:
        raise NotImplementedError

    def visit_condition_truth(self, truth) -> Truth:
        raise NotImplementedError

    def visit_condition_conjunction(self, conjunction) -> Conjunction:
        raise NotImplementedError

    def visit_condition_disjunction(self, disjunction) -> Disjunction:
        raise NotImplementedError

    def visit_condition_universal(self, universal_condition) -> UniversalCondition:
        raise NotImplementedError

    def visit_condition_existential(self, existential_condition) -> ExistentialCondition:
        raise NotImplementedError

    def visit_condition_atom(self, atom) -> Atom:
        raise NotImplementedError

    def visit_condition_negated_atom(self, negated_atom) -> NegatedAtom:
        raise NotImplementedError

    def visit_action(self, action) -> Action:
        raise NotImplementedError

    def visit_action_effect(self, effect) -> Effect:
        raise NotImplementedError

    def visit_axiom(self, axiom) -> Axiom:
        raise NotImplementedError


def contains(dumpable_object, item_name) -> bool:
    """Returns true if item_name is present in dump string of dumpable_object."""
    str_io = io.StringIO()
    with redirect_stdout(str_io):
        dumpable_object.dump()
    return item_name in str_io.getvalue()


class TaskElementErasePredicateVisitor(TaskElementVisitor):
    """Partial implementation of TaskElementVisitor interface for predicate deletion."""

    def __init__(self, predicate_name):
        self.predicate_name = predicate_name

    def visit_task(self, task):
        new_predicates = [
            predicate for predicate in task.predicates if predicate.name != self.predicate_name]

        new_init = [atom for atom in task.init if isinstance(atom, Assign)
                    or atom.predicate != self.predicate_name]

        new_goal = task.goal.accept(self)

        new_actions = []
        for action in task.actions:
            new_actions.append(action.accept(self))
        new_actions = [action for action in new_actions if
                       action.effects and not isinstance(action.precondition, Falsity)]

        new_axioms = []
        for axiom in task.axioms:
            new_axioms.append(axiom.accept(self))

        # axioms whose (head became empty OR condition became falsity) were returned as None and must be filtered out
        new_axioms = [axiom for axiom in new_axioms if axiom is not None]
        is_dummy_trigger_added = False
        trigger_id = "dummy_axiom_trigger"
        for ax in new_axioms:
            if isinstance(ax.condition, Truth):  # dummy axiom trigger needs to be created
                if not is_dummy_trigger_added:
                    dummy_axiom_trigger = Predicate(trigger_id, [])
                    new_predicates.append(dummy_axiom_trigger)
                    new_init.append(Atom(trigger_id, []))
                    is_dummy_trigger_added = True
                ax.condition = Atom(trigger_id, [])

        return Task(task.domain_name, task.task_name, task.requirements, task.types, task.objects, new_predicates,
                    task.functions, new_init, new_goal, new_actions, new_axioms, task.use_min_cost_metric)

    def visit_condition_falsity(self, falsity):
        return Falsity()

    def visit_condition_truth(self, truth):
        return Truth()

    def visit_condition_conjunction(self, conjunction):
        new_parts = []
        for part in conjunction.parts:
            new_parts.append(self.visit_condition(part))
        return Conjunction(new_parts).simplified()

    def visit_condition_disjunction(self, disjunction):
        new_parts = []
        for part in disjunction.parts:
            new_parts.append(self.visit_condition(part))
        return Disjunction(new_parts).simplified()

    def visit_condition_universal(self, universal_condition):
        new_parts = []
        for part in universal_condition.parts:
            new_parts.append(self.visit_condition(part))
        return UniversalCondition(universal_condition.parameters, new_parts).simplified()

    def visit_condition_existential(self, existential_condition):
        new_parts = []
        for part in existential_condition.parts:
            new_parts.append(self.visit_condition(part))
        return ExistentialCondition(existential_condition.parameters, new_parts).simplified()

    def visit_action(self, action):
        new_precondition = action.precondition.accept(self)

        # maybe parameters will have to be updated after predicate is deleted
        new_parameters = action.parameters
        new_num_external_parameters = action.num_external_parameters
        new_effects = []
        for effect in action.effects:
            new_effects.append(effect.accept(self))
        new_effects = [eff for eff in new_effects if
                       eff is not None and not isinstance(eff.literal, ConstantCondition) and not isinstance(
                           eff.condition, Falsity)]

        # name stays the same
        # cost stays the same
        return Action(action.name, new_parameters, new_num_external_parameters, new_precondition, new_effects,
                      action.cost)

    def visit_action_effect(self, effect):
        new_condition = effect.condition.accept(self)
        # parameters stay the same
        new_literal = effect.literal.accept(self)
        return Effect(effect.parameters, new_condition, new_literal)

    def visit_axiom(self, axiom):
        if axiom.name == self.predicate_name:  # axiom head is about to be deleted
            return None
        new_condition = axiom.condition.accept(self)
        if isinstance(new_condition, Falsity):  # axiom will never fire
            return None
        #  truth conditions are handled in visit_task
        return Axiom(axiom.name, axiom.parameters, axiom.num_external_parameters, new_condition)


class TaskElementErasePredicateTrueAtomVisitor(TaskElementErasePredicateVisitor):
    """Deletes predicates from PDDL tasks by replacing the affected atom with the truth value."""

    def visit_condition_atom(self, atom):
        if atom.predicate == self.predicate_name:
            return Truth()
        else:
            return atom

    def visit_condition_negated_atom(self, negated_atom):
        if negated_atom.predicate == self.predicate_name:
            return Falsity()
        else:
            return negated_atom


class TaskElementErasePredicateFalseAtomVisitor(TaskElementErasePredicateVisitor):
    """Deletes predicates from PDDL tasks by replacing the affected atom with the falsity value."""

    def visit_condition_atom(self, atom):
        if atom.predicate == self.predicate_name:
            return Falsity()
        else:
            return atom

    def visit_condition_negated_atom(self, negated_atom):
        if negated_atom.predicate == self.predicate_name:
            return Truth()
        else:
            return negated_atom


class TaskElementErasePredicateTrueLiteralVisitor(TaskElementErasePredicateVisitor):
    """Deletes predicates from PDDL tasks by replacing the affected literal with the truth value."""

    def visit_condition_atom(self, atom):
        if atom.predicate == self.predicate_name:
            return Truth()
        else:
            return atom

    def visit_condition_negated_atom(self, negated_atom):
        if negated_atom.predicate == self.predicate_name:
            return Truth()
        else:
            return negated_atom


class TaskElementEraseActionVisitor(TaskElementVisitor):
    """Deletes actions from PDDL tasks."""

    def __init__(self, action_name):
        self.action_name = action_name

    def visit_task(self, task):
        # filter out actions with name == self.action_name
        new_actions = [action for action in task.actions if not (
            action.name == self.action_name)]

        return Task(task.domain_name, task.task_name, task.requirements, task.types, task.objects, task.predicates,
                    task.functions, task.init, task.goal, new_actions, task.axioms, task.use_min_cost_metric)


class TaskElementEraseObjectVisitor(TaskElementVisitor):
    """Deletes objects from PDDL tasks."""

    def __init__(self, object_name):
        self.object_name = object_name

    def visit_task(self, task):
        new_objects = [
            obj for obj in task.objects if obj.name != self.object_name]
        new_init = [literal for literal in task.init if not contains(
            literal, self.object_name)]
        new_actions = [action for action in task.actions if not contains(
            action, self.object_name)]
        new_goal = task.goal.accept(self)

        return Task(task.domain_name, task.task_name, task.requirements, task.types, new_objects, task.predicates,
                    task.functions, new_init, new_goal, new_actions, task.axioms, task.use_min_cost_metric)

    def visit_condition_falsity(self, falsity) -> Falsity:
        return Falsity()

    def visit_condition_truth(self, truth) -> Truth:
        return Truth()

    def visit_condition_conjunction(self, conjunction) -> Conjunction:
        new_parts = []
        for part in conjunction.parts:
            new_parts.append(part.accept(self))
        new_parts = [part for part in new_parts if part is not None]
        return Conjunction(new_parts).simplified()

    def visit_condition_disjunction(self, disjunction) -> Disjunction:
        new_parts = []
        for part in disjunction.parts:
            new_parts.append(part.accept(self))
        new_parts = [part for part in new_parts if part is not None]
        return Disjunction(new_parts).simplified()

    def visit_condition_universal(self, universal_condition) -> UniversalCondition:
        new_parts = []
        for part in universal_condition.parts:
            new_parts.append(part.accept(self))
        new_parts = [part for part in new_parts if part is not None]
        return UniversalCondition(universal_condition.parameters, new_parts).simplified()

    def visit_condition_existential(self, existential_condition) -> ExistentialCondition:
        new_parts = []
        for part in existential_condition.parts:
            new_parts.append(part.accept(self))
        new_parts = [part for part in new_parts if part is not None]
        return ExistentialCondition(existential_condition.parameters, new_parts).simplified()

    def visit_condition_atom(self, atom) -> Atom:
        return Truth() if contains(atom, self.object_name) else atom

    def visit_condition_negated_atom(self, negated_atom) -> NegatedAtom:
        return Falsity() if contains(negated_atom, self.object_name) else negated_atom



from machetli.planning.downward_lib.pddl.task_element import TaskElement


class Predicate(TaskElement):

    def accept(self, visitor):
        return visitor.visit_predicate(self)

    def __init__(self, name, arguments):
        self.name = name
        self.arguments = arguments

    def __str__(self):
        return "%s(%s)" % (self.name, ", ".join(map(str, self.arguments)))

    def get_arity(self):
        return len(self.arguments)



import copy

from . import conditions, TaskElement


class Action(TaskElement):
    def accept(self, visitor):
        return visitor.visit_action(self)

    def __init__(self, name, parameters, num_external_parameters,
                 precondition, effects, cost):
        assert 0 <= num_external_parameters <= len(parameters)
        self.name = name
        self.parameters = parameters
        # num_external_parameters denotes how many of the parameters
        # are "external", i.e., should be part of the grounded action
        # name. Usually all parameters are external, but "invisible"
        # parameters can be created when compiling away existential
        # quantifiers in conditions.
        self.num_external_parameters = num_external_parameters
        self.precondition = precondition
        self.effects = effects
        self.cost = cost
        self.uniquify_variables()  # TODO: uniquify variables in cost?

    def __repr__(self):
        return "<Action %r at %#x>" % (self.name, id(self))

    def dump(self):
        print("%s(%s)" % (self.name, ", ".join(map(str, self.parameters))))
        print("Precondition:")
        self.precondition.dump()
        print("Effects:")
        for eff in self.effects:
            eff.dump()
        print("Cost:")
        if (self.cost):
            self.cost.dump()
        else:
            print("  None")

    def uniquify_variables(self):
        self.type_map = {par.name: par.type_name for par in self.parameters}
        self.precondition = self.precondition.uniquify_variables(self.type_map)
        for effect in self.effects:
            effect.uniquify_variables(self.type_map)

    def relaxed(self):
        new_effects = []
        for eff in self.effects:
            relaxed_eff = eff.relaxed()
            if relaxed_eff:
                new_effects.append(relaxed_eff)
        return Action(self.name, self.parameters, self.num_external_parameters,
                      self.precondition.relaxed().simplified(),
                      new_effects)

    def untyped(self):
        # We do not actually remove the types from the parameter lists,
        # just additionally incorporate them into the conditions.
        # Maybe not very nice.
        result = copy.copy(self)
        parameter_atoms = [par.to_untyped_strips() for par in self.parameters]
        new_precondition = self.precondition.untyped()
        result.precondition = conditions.Conjunction(parameter_atoms + [new_precondition])
        result.effects = [eff.untyped() for eff in self.effects]
        return result

    def instantiate(self, var_mapping, init_facts, init_assignments,
                    fluent_facts, objects_by_type, metric):
        """Return a PropositionalAction which corresponds to the instantiation of
        this action with the arguments in var_mapping. Only fluent parts of the
        conditions (those in fluent_facts) are included. init_facts are evaluated
        while instantiating.
        Precondition and effect conditions must be normalized for this to work.
        Returns None if var_mapping does not correspond to a valid instantiation
        (because it has impossible preconditions or an empty effect list.)"""
        arg_list = [var_mapping[par.name]
                    for par in self.parameters[:self.num_external_parameters]]
        name = "(%s %s)" % (self.name, " ".join(arg_list))

        precondition = []
        try:
            self.precondition.instantiate(var_mapping, init_facts,
                                          fluent_facts, precondition)
        except conditions.Impossible:
            return None
        effects = []
        for eff in self.effects:
            eff.instantiate(var_mapping, init_facts, fluent_facts,
                            objects_by_type, effects)
        if effects:
            if metric:
                if self.cost is None:
                    cost = 0
                else:
                    cost = int(self.cost.instantiate(
                        var_mapping, init_assignments).expression.value)
            else:
                cost = 1
            return PropositionalAction(name, precondition, effects, cost)
        else:
            return None


class PropositionalAction:
    def __init__(self, name, precondition, effects, cost):
        self.name = name
        self.precondition = precondition
        self.add_effects = []
        self.del_effects = []
        for condition, effect in effects:
            if not effect.negated:
                self.add_effects.append((condition, effect))
        # Warning: This is O(N^2), could be turned into O(N).
        # But that might actually harm performance, since there are
        # usually few effects.
        # TODO: Measure this in critical domains, then use sets if acceptable.
        for condition, effect in effects:
            if effect.negated and (condition, effect.negate()) not in self.add_effects:
                self.del_effects.append((condition, effect.negate()))
        self.cost = cost

    def __repr__(self):
        return "<PropositionalAction %r at %#x>" % (self.name, id(self))

    def dump(self):
        print(self.name)
        for fact in self.precondition:
            print("PRE: %s" % fact)
        for cond, fact in self.add_effects:
            print("ADD: %s -> %s" % (", ".join(map(str, cond)), fact))
        for cond, fact in self.del_effects:
            print("DEL: %s -> %s" % (", ".join(map(str, cond)), fact))
        print("cost:", self.cost)



#! /usr/bin/env python3

import copy

import pddl

class ConditionProxy:
    def clone_owner(self):
        clone = copy.copy(self)
        clone.owner = copy.copy(clone.owner)
        return clone

class PreconditionProxy(ConditionProxy):
    def __init__(self, action):
        self.owner = action
        self.condition = action.precondition
    def set(self, new_condition):
        self.owner.precondition = self.condition = new_condition
    def register_owner(self, task):
        task.actions.append(self.owner)
    def delete_owner(self, task):
        task.actions.remove(self.owner)
    def build_rules(self, rules):
        action = self.owner
        rule_head = get_action_predicate(action)
        rule_body = condition_to_rule_body(action.parameters, self.condition)
        rules.append((rule_body, rule_head))
    def get_type_map(self):
        return self.owner.type_map

class EffectConditionProxy(ConditionProxy):
    def __init__(self, action, effect):
        self.action = action
        self.owner = effect
        self.condition = effect.condition
    def set(self, new_condition):
        self.owner.condition = self.condition = new_condition
    def register_owner(self, task):
        self.action.effects.append(self.owner)
    def delete_owner(self, task):
        self.action.effects.remove(self.owner)
    def build_rules(self, rules):
        effect = self.owner
        rule_head = effect.literal
        if not rule_head.negated:
            rule_body = [get_action_predicate(self.action)]
            rule_body += condition_to_rule_body([], self.condition)
            rules.append((rule_body, rule_head))
    def get_type_map(self):
        return self.action.type_map

class AxiomConditionProxy(ConditionProxy):
    def __init__(self, axiom):
        self.owner = axiom
        self.condition = axiom.condition
    def set(self, new_condition):
        self.owner.condition = self.condition = new_condition
    def register_owner(self, task):
        task.axioms.append(self.owner)
    def delete_owner(self, task):
        task.axioms.remove(self.owner)
    def build_rules(self, rules):
        axiom = self.owner
        app_rule_head = get_axiom_predicate(axiom)
        app_rule_body = condition_to_rule_body(axiom.parameters, self.condition)
        rules.append((app_rule_body, app_rule_head))
        params = axiom.parameters[:axiom.num_external_parameters]
        eff_rule_head = pddl.Atom(axiom.name, [par.name for par in params])
        eff_rule_body = [app_rule_head]
        rules.append((eff_rule_body, eff_rule_head))
    def get_type_map(self):
        return self.owner.type_map

class GoalConditionProxy(ConditionProxy):
    def __init__(self, task):
        self.owner = task
        self.condition = task.goal
    def set(self, new_condition):
        self.owner.goal = self.condition = new_condition
    def register_owner(self, task):
        # this assertion should never trigger, because disjunctive
        # goals are now implemented with axioms
        # (see substitute_complicated_goal)
        assert False, "Disjunctive goals not (yet) implemented."
    def delete_owner(self, task):
        # this assertion should never trigger, because disjunctive
        # goals are now implemented with axioms
        # (see substitute_complicated_goal)
        assert False, "Disjunctive goals not (yet) implemented."
    def build_rules(self, rules):
        rule_head = pddl.Atom("@goal-reachable", [])
        rule_body = condition_to_rule_body([], self.condition)
        rules.append((rule_body, rule_head))
    def get_type_map(self):
        # HACK!
        # Method uniquify_variables HAS already been called (which is good).
        # We call it here again for its SIDE EFFECT of collecting the type_map
        # (which is bad). Having "top-level conditions" (currently, only goal
        # conditions, but might also include safety conditions and similar)
        # contained in a separate wrapper class that stores a type map might
        # be a better design.
        type_map = {}
        self.condition.uniquify_variables(type_map)
        return type_map

def get_action_predicate(action):
    name = action
    variables = [par.name for par in action.parameters]
    if isinstance(action.precondition, pddl.ExistentialCondition):
        variables += [par.name for par in action.precondition.parameters]
    return pddl.Atom(name, variables)

def get_axiom_predicate(axiom):
    name = axiom
    variables = [par.name for par in axiom.parameters]
    if isinstance(axiom.condition, pddl.ExistentialCondition):
        variables += [par.name for par in axiom.condition.parameters]
    return pddl.Atom(name, variables)

def all_conditions(task):
    for action in task.actions:
        yield PreconditionProxy(action)
        for effect in action.effects:
            yield EffectConditionProxy(action, effect)
    for axiom in task.axioms:
        yield AxiomConditionProxy(axiom)
    yield GoalConditionProxy(task)

# [1] Remove universal quantifications from conditions.
#
# Replace, in a top-down fashion, <forall(vars, phi)> by <not(not-all-phi)>,
# where <not-all-phi> is a new axiom.
#
# <not-all-phi> is defined as <not(forall(vars,phi))>, which is of course
# translated to NNF. The parameters of the new axioms are exactly the free
# variables of <forall(vars, phi)>.

def remove_universal_quantifiers(task):
    def recurse(condition):
        # Uses new_axioms_by_condition and type_map from surrounding scope.
        if isinstance(condition, pddl.UniversalCondition):
            axiom_condition = condition.negate()
            parameters = sorted(axiom_condition.free_variables())
            typed_parameters = tuple(pddl.TypedObject(v, type_map[v]) for v in parameters)
            axiom = new_axioms_by_condition.get((axiom_condition, typed_parameters))
            if not axiom:
                condition = recurse(axiom_condition)
                axiom = task.add_axiom(list(typed_parameters), condition)
                new_axioms_by_condition[(condition, typed_parameters)] = axiom
            return pddl.NegatedAtom(axiom.name, parameters)
        else:
            new_parts = [recurse(part) for part in condition.parts]
            return condition.change_parts(new_parts)

    new_axioms_by_condition = {}
    for proxy in tuple(all_conditions(task)):
        # Cannot use generator because we add new axioms on the fly.
        if proxy.condition.has_universal_part():
            type_map = proxy.get_type_map()
            proxy.set(recurse(proxy.condition))


# [2] Pull disjunctions to the root of the condition.
#
# After removing universal quantifiers, the (k-ary generalization of the)
# following rules suffice for doing that:
# (1) or(phi, or(psi, psi'))      ==  or(phi, psi, psi')
# (2) exists(vars, or(phi, psi))  ==  or(exists(vars, phi), exists(vars, psi))
# (3) and(phi, or(psi, psi'))     ==  or(and(phi, psi), and(phi, psi'))
def build_DNF(task):
    def recurse(condition):
        disjunctive_parts = []
        other_parts = []
        for part in condition.parts:
            part = recurse(part)
            if isinstance(part, pddl.Disjunction):
                disjunctive_parts.append(part)
            else:
                other_parts.append(part)
        if not disjunctive_parts:
            return condition

        # Rule (1): Associativity of disjunction.
        if isinstance(condition, pddl.Disjunction):
            result_parts = other_parts
            for part in disjunctive_parts:
                result_parts.extend(part.parts)
            return pddl.Disjunction(result_parts)

        # Rule (2): Distributivity disjunction/existential quantification.
        if isinstance(condition, pddl.ExistentialCondition):
            parameters = condition.parameters
            result_parts = [pddl.ExistentialCondition(parameters, (part,))
                            for part in disjunctive_parts[0].parts]
            return pddl.Disjunction(result_parts)

        # Rule (3): Distributivity disjunction/conjunction.
        assert isinstance(condition, pddl.Conjunction)
        result_parts = [pddl.Conjunction(other_parts)]
        while disjunctive_parts:
            previous_result_parts = result_parts
            result_parts = []
            parts_to_distribute = disjunctive_parts.pop().parts
            for part1 in previous_result_parts:
                for part2 in parts_to_distribute:
                    result_parts.append(pddl.Conjunction((part1, part2)))
        return pddl.Disjunction(result_parts)

    for proxy in all_conditions(task):
        if proxy.condition.has_disjunction():
            proxy.set(recurse(proxy.condition).simplified())

# [3] Split conditions at the outermost disjunction.
def split_disjunctions(task):
    for proxy in tuple(all_conditions(task)):
        # Cannot use generator directly because we add/delete entries.
        if isinstance(proxy.condition, pddl.Disjunction):
            for part in proxy.condition.parts:
                new_proxy = proxy.clone_owner()
                new_proxy.set(part)
                new_proxy.register_owner(task)
            proxy.delete_owner(task)

# [4] Pull existential quantifiers out of conjunctions and group them.
#
# After removing universal quantifiers and creating the disjunctive form,
# only the following (representatives of) rules are needed:
# (1) exists(vars, exists(vars', phi))  ==  exists(vars + vars', phi)
# (2) and(phi, exists(vars, psi))       ==  exists(vars, and(phi, psi)),
#       if var does not occur in phi as a free variable.
def move_existential_quantifiers(task):
    def recurse(condition):
        existential_parts = []
        other_parts = []
        for part in condition.parts:
            part = recurse(part)
            if isinstance(part, pddl.ExistentialCondition):
                existential_parts.append(part)
            else:
                other_parts.append(part)
        if not existential_parts:
            return condition

        # Rule (1): Combine nested quantifiers.
        if isinstance(condition, pddl.ExistentialCondition):
            new_parameters = condition.parameters + existential_parts[0].parameters
            new_parts = existential_parts[0].parts
            return pddl.ExistentialCondition(new_parameters, new_parts)

        # Rule (2): Pull quantifiers out of conjunctions.
        assert isinstance(condition, pddl.Conjunction)
        new_parameters = []
        new_conjunction_parts = other_parts
        for part in existential_parts:
            new_parameters += part.parameters
            new_conjunction_parts += part.parts
        new_conjunction = pddl.Conjunction(new_conjunction_parts)
        return pddl.ExistentialCondition(new_parameters, (new_conjunction,))

    for proxy in all_conditions(task):
        if proxy.condition.has_existential_part():
            proxy.set(recurse(proxy.condition).simplified())


# [5a] Drop existential quantifiers from axioms, turning them
#      into parameters.

def eliminate_existential_quantifiers_from_axioms(task):
    # Note: This is very redundant with the corresponding method for
    # actions and could easily be merged if axioms and actions were
    # unified.
    for axiom in task.axioms:
        precond = axiom.condition
        if isinstance(precond, pddl.ExistentialCondition):
            # Copy parameter list, since it can be shared with
            # parameter lists of other versions of this axiom (e.g.
            # created when splitting up disjunctive preconditions).
            axiom.parameters = list(axiom.parameters)
            axiom.parameters.extend(precond.parameters)
            axiom.condition = precond.parts[0]


# [5b] Drop existential quantifiers from action preconditions,
#      turning them into action parameters (that don't form part of the
#      name of the action).

def eliminate_existential_quantifiers_from_preconditions(task):
    for action in task.actions:
        precond = action.precondition
        if isinstance(precond, pddl.ExistentialCondition):
            # Copy parameter list, since it can be shared with
            # parameter lists of other versions of this action (e.g.
            # created when splitting up disjunctive preconditions).
            action.parameters = list(action.parameters)
            action.parameters.extend(precond.parameters)
            action.precondition = precond.parts[0]

# [5c] Eliminate existential quantifiers from effect conditions
#
# For effect conditions, we replace "when exists(x, phi) then e" with
# "forall(x): when phi then e.
def eliminate_existential_quantifiers_from_conditional_effects(task):
    for action in task.actions:
        for effect in action.effects:
            condition = effect.condition
            if isinstance(condition, pddl.ExistentialCondition):
                effect.parameters = list(effect.parameters)
                effect.parameters.extend(condition.parameters)
                effect.condition = condition.parts[0]

def substitute_complicated_goal(task):
    goal = task.goal
    if isinstance(goal, pddl.Literal):
        return
    elif isinstance(goal, pddl.Conjunction):
        for item in goal.parts:
            if not isinstance(item, pddl.Literal):
                break
        else:
            return
    new_axiom = task.add_axiom([], goal)
    task.goal = pddl.Atom(new_axiom.name, new_axiom.parameters)

# Combine Steps [1], [2], [3], [4], [5] and do some additional verification
# that the task makes sense.

def normalize(task):
    remove_universal_quantifiers(task)
    substitute_complicated_goal(task)
    build_DNF(task)
    split_disjunctions(task)
    move_existential_quantifiers(task)
    eliminate_existential_quantifiers_from_axioms(task)
    eliminate_existential_quantifiers_from_preconditions(task)
    eliminate_existential_quantifiers_from_conditional_effects(task)

    verify_axiom_predicates(task)

def verify_axiom_predicates(task):
    # Verify that derived predicates are not used in :init or
    # action effects.
    axiom_names = set()
    for axiom in task.axioms:
        axiom_names.add(axiom.name)

    for fact in task.init:
        # Note that task.init can contain the assignment to (total-cost)
        # in addition to regular atoms.
        if getattr(fact, "predicate", None) in axiom_names:
            raise SystemExit(
                "error: derived predicate %r appears in :init fact '%s'" %
                (fact.predicate, fact))

    for action in task.actions:
        for effect in action.effects:
            if effect.literal.predicate in axiom_names:
                raise SystemExit(
                    "error: derived predicate %r appears in effect of action %r" %
                    (effect.literal.predicate, action.name))


# [6] Build rules for exploration component.
def build_exploration_rules(task):
    result = []
    for proxy in all_conditions(task):
        proxy.build_rules(result)
    return result

def condition_to_rule_body(parameters, condition):
    result = []
    for par in parameters:
        result.append(par.get_atom())
    if not isinstance(condition, pddl.Truth):
        if isinstance(condition, pddl.ExistentialCondition):
            for par in condition.parameters:
                result.append(par.get_atom())
            condition = condition.parts[0]
        if isinstance(condition, pddl.Conjunction):
            parts = condition.parts
        else:
            parts = (condition,)
        for part in parts:
            if isinstance(part, pddl.Falsity):
                # Use an atom in the body that is always false because
                # it is not initially true and doesn't occur in the
                # head of any rule.
                return [pddl.Atom("@always-false", [])]
            assert isinstance(part, pddl.Literal), "Condition not normalized: %r" % part
            if not part.negated:
                result.append(part)
    return result

if __name__ == "__main__":
    import pddl_parser
    task = pddl_parser.open()
    normalize(task)
    task.dump()



from setuptools import setup, find_packages

with open("README.rst", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name="machetli",
    version="0.2",
    description="Locate bugs in your program",
    long_description=long_description,
    long_description_content_type="text/x-rst",
    author="Lucas Galery Käser",
    author_email="lucas.galerykaeser@gmail.com",
    url="https://github.com/aibasel/machetli",
    license="GPL3+",
    project_urls={
        "Documentation": "https://minimizer.readthedocs.io/",
    },
    classifiers=[
        "Environment :: Console",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Topic :: Scientific/Engineering",
    ],
    packages=find_packages(),
    python_requires=">=3.7",
    include_package_data=True,
    package_data={
        "machetli": ["grid/slurm-array-job.template"],
    },
)



"""
This file is derived from ``parser.py`` of Lab (<https://lab.readthedocs.io>).

Our *_ContentParser* is basically the *_FileParser* from Lab, except
that it has an *initialize* function to define the content to be parsed
instead of the *load_file* function.

The *Parser* class itself is adapted so that a pattern or function is
added for each *cmd_name* instead of the *file*.
"""

from collections import defaultdict
import logging
import re

from machetli import tools


def _get_pattern_flags(s):
    flags = 0
    for char in s:
        try:
            flags |= getattr(re, char)
        except AttributeError:
            logging.critical(f"Unknown pattern flag: {char}")
    return flags


class _Pattern:
    def __init__(self, attribute, regex, required, type_, flags):
        self.attribute = attribute
        self.type_ = type_
        self.required = required
        self.group = 1

        flags = _get_pattern_flags(flags)
        self.regex = re.compile(regex, flags)

    def search(self, content, cmd_name):
        found_props = {}
        match = self.regex.search(content)
        if match:
            try:
                value = match.group(self.group)
            except IndexError:
                logging.error(
                    f"Attribute {self.attribute} not found for pattern "
                    f"{self} in content of command {cmd_name}."
                )
            else:
                value = self.type_(value)
                found_props[self.attribute] = value
        return found_props

    def __str__(self):
        return self.regex.pattern


class _ContentParser:
    """
    Private class that parses a given file according to the added patterns
    and functions.
    """

    def __init__(self):
        self.patterns = []
        self.functions = []

    def add_pattern(self, pattern):
        self.patterns.append(pattern)

    def add_function(self, function):
        self.functions.append(function)

    def search_patterns(self, content, cmd_name):
        assert content is not None
        found_props = {}
        for pattern in self.patterns:
            found_props.update(pattern.search(content, cmd_name))
        return found_props

    def apply_functions(self, props, content):
        assert content is not None
        for function in self.functions:
            function(content, props)


class Parser:
    """
    Parse stdout and stderr strings.
    """
    def __init__(self):
        tools.configure_logging()
        self.content_parsers = defaultdict(_ContentParser)

    def add_pattern(self, attribute, regex, cmd_names, type=int, flags=""):
        """
        Look for *regex* in stdout and stderr of the executed runs
        with names *cmd_names* and cast what is found in brackets to *type*.
        
        Store the parsing result of this pattern under the name
        *attribute* in the properties dictionary returned by
        :meth:`parse(cmd_name, content) <machetli.parser.Parser.parse>`.

        *flags* must be a string of Python regular expression flags (see
        https://docs.python.org/3/library/re.html). E.g., ``flags="M"``
        lets "^" and "$" match at the beginning and end of each line,
        respectively.

        Usage example:

        .. code-block:: python
        
            parser = Parser()

            parser.add_pattern("translator_facts",
                   r"Translator facts: (\d+)", "amazing_run")
        """
        if type == bool:
            logging.warning(
                "Casting any non-empty string to boolean will always "
                "evaluate to true. Are you sure you want to use type=bool?"
            )
        pattern = _Pattern(attribute, regex, required=False, type_=type,
                           flags=flags)
        for name in tools.make_list(cmd_names):
            self.content_parsers[name].add_pattern(pattern)

    def add_function(self, function, cmd_names):
        """
        Add *function* to parser which is called on the content strings
        of the executed runs *cmd_names*. *cmd_names* can be used for
        single arguments and for argument lists.

        Functions are applied **after** all patterns have been evaluated.

        The function is passed the content strings and the properties
        dictionary. It must manipulate the passed properties dictionary.
        The return value is ignored.

        Usage example:

        .. code-block:: python

            parser = Parser()

            def facts_tracker(content, props):
                props["translator_facts"] =
                    re.findall(r"Translator facts: (\d+)", content)

            parser.add_function(facts_tracker, ["amazing_run", "superb_run"])
        """
        for name in tools.make_list(cmd_names):
            self.content_parsers[name].add_function(function)

    def parse(self, cmd_name, content):
        """
        Search all patterns and apply all functions to *content* of run
        *cmd_name*.
        """
        self.props = dict()

        for name, content_parser in list(self.content_parsers.items()):
            if name == cmd_name:
                self.props.update(
                    content_parser.search_patterns(content, cmd_name))

        for name, content_parser in list(self.content_parsers.items()):
            if name == cmd_name:
                content_parser.apply_functions(self.props, content)

        return self.props


if __name__ == "__main__":
    import pprint
    parser = Parser()
    parser.add_pattern(attribute="attr", regex=r"(world)",
                       cmd_names="test", type=str)
    result = parser.parse(cmd_name="test", content="Hello world!")
    pprint.pprint(result)



import importlib
import logging
import platform
import sys

from machetli.tools import read_state

# https://stackoverflow.com/questions/67631/how-to-import-a-module-given-the-full-path/50395128#50395128
def _import_evaluator(module_name, evaluator_path):
    spec = importlib.util.spec_from_file_location(module_name, evaluator_path)
    module = importlib.util.module_from_spec(spec)
    sys.modules[spec.name] = module 
    spec.loader.exec_module(module)
    return module


def is_evaluator_successful(evaluator_path, state):
    module = _import_evaluator("custom_evaluator", evaluator_path)
    return module.evaluate(state)


if __name__ == "__main__":
    if len(sys.argv) != 3:
        logging.critical(f"Expected two arguments to machetli.evaluator but got {len(sys.argv)}.")
    evaluator_path = sys.argv[1]
    state_filename = sys.argv[2]
    logging.info(f"Running evaluator '{evaluator_path}' for state '{state_filename}' on node: {platform.node()}.")

    state = read_state(state_filename, 5, 2)
    if is_evaluator_successful(evaluator_path, state):
        sys.exit(0)
    else:
        sys.exit(1)



# import options

from . import lisp_parser
from . import parsing_functions
# from .. import options

file_open = open


def parse_pddl_file(type, filename):
    try:
        # The builtin open function is shadowed by this module's open function.
        # We use the Latin-1 encoding (which allows a superset of ASCII, of the
        # Latin-* encodings and of UTF-8) to allow special characters in
        # comments. In all other parts, we later validate that only ASCII is
        # used.
        return lisp_parser.parse_nested_list(file_open(filename,
                                                       encoding='ISO-8859-1'))
    except OSError as e:
        raise SystemExit("Error: Could not read file: %s\nReason: %s." %
                         (e.filename, e))
    except lisp_parser.ParseError as e:
        raise SystemExit("Error: Could not parse %s file: %s\nReason: %s." %
                         (type, filename, e))


def open(domain_filename=None, task_filename=None):
    task_filename = task_filename #or options.task
    domain_filename = domain_filename #or options.domain

    domain_pddl = parse_pddl_file("domain", domain_filename)
    task_pddl = parse_pddl_file("task", task_filename)

    return parsing_functions.parse_task(domain_pddl, task_pddl)



"""
This file is derived from ``tools.py`` of Lab (<https://lab.readthedocs.io>).
Functions and classes that are not needed for this project were removed.
"""
import logging
import os
import pickle
import pprint
import random
import sys
import time


DEFAULT_ENCODING = "utf-8"


def get_string(s):
    if isinstance(s, bytes):
        return s.decode(DEFAULT_ENCODING)
    else:
        raise ValueError("tools.get_string() only accepts byte strings")


def get_script_path():
    """Get absolute path to main script."""
    return os.path.abspath(sys.argv[0])


def get_python_executable():
    return sys.executable or "python"


def configure_logging(level=logging.INFO):
    # Python adds a default handler if some log is written before this
    # function is called. We therefore remove all handlers that have
    # been added automatically.
    root_logger = logging.getLogger("")
    for handler in root_logger.handlers:
        root_logger.removeHandler(handler)

    class ErrorAbortHandler(logging.StreamHandler):
        """
        Logging handler that exits when a critical error is encountered.
        """

        def emit(self, record):
            logging.StreamHandler.emit(self, record)
            if record.levelno >= logging.CRITICAL:
                sys.exit("aborting")

    class StdoutFilter(logging.Filter):
        def filter(self, record):
            return record.levelno <= logging.WARNING

    class StderrFilter(logging.Filter):
        def filter(self, record):
            return record.levelno > logging.WARNING

    formatter = logging.Formatter("%(asctime)-s %(levelname)-8s %(message)s")

    stdout_handler = logging.StreamHandler(sys.stdout)
    stdout_handler.setFormatter(formatter)
    stdout_handler.addFilter(StdoutFilter())

    stderr_handler = ErrorAbortHandler(sys.stderr)
    stderr_handler.setFormatter(formatter)
    stderr_handler.addFilter(StderrFilter())

    root_logger.addHandler(stdout_handler)
    root_logger.addHandler(stderr_handler)
    root_logger.setLevel(level)

# TODO: only used by a deprecated parser function. will be removed.
def make_list(value):
    if value is None:
        return []
    elif isinstance(value, list):
        return value[:]
    elif isinstance(value, (tuple, set)):
        return list(value)
    else:
        return [value]


def makedirs(path):
    """
    os.makedirs() variant that doesn't complain if the path already exists.
    """
    try:
        os.makedirs(path)
    except OSError:
        # Directory probably already exists.
        pass


def write_state(state, file_path):
    with open(file_path, "wb") as state_file:
        pickle.dump(state, state_file)


def read_state(file_path, wait_time, repetitions):
    for _ in range(repetitions):
        time.sleep(wait_time * random.random())
        if os.path.exists(file_path):
            with open(file_path, "rb") as state_file:
                return pickle.load(state_file)
    else:
        logging.critical(f"Could not find file '{filename}' after {repetitions} attempts.")


class SubmissionError(Exception):
    def __init__(self, cpe):
        print(cpe)
        self.returncode = cpe.returncode
        self.cmd = cpe.cmd
        self.output = cpe.output
        self.stdout = cpe.stdout
        self.stderr = cpe.stderr

    def __str__(self):
        return f"""
                Error during job submission:
                Submission command: {self.cmd}
                Returncode: {self.returncode}
                Output: {self.output}
                Captured stdout: {self.stdout}
                Captured stderr: {self.stderr}"""

    def warn(self):
        logging.warning(f"The following batch submission failed but is "
                        f"ignored: {self}")

    def warn_abort(self):
        logging.warning(
            f"Task order cannot be kept because the following batch "
            f"submission failed: {self} Aborting search.")


class TaskError(Exception):
    def __init__(self, critical_tasks):
        self.critical_tasks = critical_tasks
        self.indices_critical = [int(parts[1]) for parts in (
            task_id.split("_") for task_id in self.critical_tasks)]

    def __repr__(self):
        return pprint.pformat(self.critical_tasks)

    def remove_critical_tasks(self, job):
        """Remove tasks from job that entered a critical state."""
        job["tasks"] = [t for i, t in enumerate(
            job["tasks"]) if i not in self.indices_critical]
        logging.warning(
            f"Some tasks from job {job['id']} entered a critical "
            f"state but the search is continued.")

    def remove_tasks_after_first_critical(self, job):
        """
        Remove all tasks from job after the first one that entered a
        critical state.
        """
        first_failed = self.indices_critical[0]
        job["tasks"] = job["tasks"][:first_failed]
        if not job["tasks"]:
            logging.warning("Since the first task failed, the order "
                            "cannot be kept. Aborting search.")
        else:
            logging.warning(
                f"At least one task from job {job['id']} entered a "
                f"critical state: {self} The tasks before the first "
                f"critical one are still considered.")


class PollingError(Exception):
    def __init__(self, job_id):
        self.job_id = job_id

    def warn_abort(self):
        logging.error(f"Polling job {self.job_id} caused an error. "
                      f"Aborting search.")

